import zmq
from pynq import DefaultIP, buffer
import numpy as np
from logging import getLogger
import threading
import queue
from collections import namedtuple

from ..mkidpynq import PHOTON_DTYPE as _PHOTON_DTYPE
from ..system_parameters import LOWPASSED_IQ_SAMPLE_RATE, PHOTON_POSTAGE_WINDOW_LENGTH
from ..interrupts import ThreadedPLInterruptManager
from ..mkidpynq import unpack_photons

class PhotonTrigger(DefaultIP):
    bindto = ['mazinlab:mkidgen3:trigger:0.4', 'mazinlab:mkidgen3:fake_trigger:0.1']
    THRESHOFF_SLICE = slice(0x1000 // 4, 0x1000 // 4 + 1024)
    StatusTuple = namedtuple('PhotonTriggerStatus', ('thresholds', 'holdoffs'))

    def __init__(self, description):
        """
        The core uses an array of 512 values, each consisting of 4 pairs of values packed into a 64bit dword.
        These pairs of values consist of the threshold (a byte treated as an ap_fixed<8,1> number for comparison
        against the ap_fixed<16,1> phase) and a holdoff value (a byte that specified the number of samples over which to
        pick the minimal phase for a photon).

        The holdoff value is in units of the phase stream sample rate (nominally 1MHz). It must not exceed 254 as the
        driver needs to add 1 to the requested value internally. It is expected that a value below ~7 will cause
        aberrant and unpredictable behavior in the full design even if the core itself supports it

        Although the core will issue information about photon overflow as a stream it does not maintain a register
        for it and, as such, this driver does not provide information in that regard.

        x[0] consists of (from low to high byte) thresh0 hoff0 thresh1 hoff1 thresh2 hoff2 thresh3 hoff3
        x[511] consists of (from low to high byte) thresh2044 hoff2044 ... thresh2047 hoff2047

        control
        0x1000 ~
        0x1fff : Memory 'threshoffs' (512 * 64b)
                 Word 2n   : bit [31:0] - threshoffs[n][31: 0]
                 Word 2n+1 : bit [31:0] - threshoffs[n][63:32]
        """
        super().__init__(description=description)

    def configuration(self, scaled=True):
        """
        Return named tuple of thresholds (in raw format or scaled format) and holdoffs assigned to each of the
        2048 resonator channels
        """
        x = np.frombuffer(np.array(self.mmio.array[self.THRESHOFF_SLICE]), dtype=np.uint8)
        return self.StatusTuple(thresholds=x[::2].astype(np.int8) / 128 if scaled else x[::2].astype(np.int8),
                                holdoffs=x[1::2])

    def configure(self, thresholds=None, holdoffs=None):
        """Thresholds shall be floating point numbers in [-1,1) and will be converted to ap_fixed<8,0>"""
        if thresholds is None and holdoffs is None:
            return

        _thresholds, _holdoffs = None, None
        if thresholds is not None:
            thresholds = (np.asarray(thresholds).clip(-1, 1) * 128).round().astype(int).clip(-128, 127)
            if len(thresholds) != 2048:
                raise ValueError('len(thresholds)!=2048')
        else:
            _thresholds, _holdoffs = self.configuration(scaled=False)

        if holdoffs is not None:
            holdoffs = np.asarray(holdoffs).astype(int).clip(0, 254)
            if len(holdoffs) != 2048:
                raise ValueError('len(holdoffs)!=2048')
        elif _holdoffs is None:
            _thresholds, _holdeffos = self.configuration(scaled=False)

        if thresholds is None:
            thresholds = _thresholds
        if holdoffs is None:
            holdoffs = _holdoffs + 1

        assert thresholds.size == holdoffs.size

        data = np.zeros((2048, 2), dtype=np.uint8)
        data[:, 0] = thresholds.astype(np.uint8)
        data[:, 1] = holdoffs
        self.mmio.array[self.THRESHOFF_SLICE] = np.frombuffer(data, dtype=np.uint32)


class PhotonPostageFilter(DefaultIP):
    ADDR_MONITOR_CHAN = tuple(range(0x10, 0x49, 0x8))
    bindto = ['mazinlab:mkidgen3:postage_filter:0.2', 'mazinlab:mkidgen3:postage_filter_w_interconn:0.1']

    def __init__(self, description):
        """
        The core watches for trigger events on 8 resonator channels and forwards IQ snippets around trigger events on
        for capture by other parts of the firmware. It is configured with 8 resonator channel values for monitoring.

        control
        0x10 : Data signal of monitor_0
               bit 11~0 - monitor_0[11:0] (Read/Write)
               others   - reserved
        0x18, 0x20, 0x28,0x30,0x38,0x40 0x48 : Data signal of monitor_1 - 7

        """
        super().__init__(description=description)

    @property
    def monitor_channels(self):
        return tuple(self.read(a) & 0x7ff for a in self.ADDR_MONITOR_CHAN)

    def configure(self, monitor_channels=None):
        """ monitor_channels shall be 8 integers in [0,2047] """
        if monitor_channels is None:
            return
        monitor_channels = np.asarray(monitor_channels[:8], dtype=np.uint16).clip(0, 2047)
        for a, c in zip(self.ADDR_MONITOR_CHAN, monitor_channels):
            self.write(a, int(c))


class PhotonPostageMAXI(DefaultIP):
    POSTAGE_BUFFER_LEN = 1000 * 8  # Must be POSTAGE_BUFSIZE from the HLS
    N_CAPDATA = PHOTON_POSTAGE_WINDOW_LENGTH  # Must be 1 less than the HLS value
    MAX_CAPTURE_TIME_S = POSTAGE_BUFFER_LEN / 8 * (N_CAPDATA + 1) / LOWPASSED_IQ_SAMPLE_RATE
    bindto = ['mazinlab:mkidgen3:postage_maxi:0.2']

    def __init__(self, description):
        """
        The core captures up to some number of Photon events to a pynq buffer. The resonators for monitoring are
        configured by PhotonPostageFilter.

        // control
        // 0x00 : Control signals
        //        bit 0  - ap_start (Read/Write/COH)
        //        bit 1  - ap_done (Read/COR)
        //        bit 2  - ap_idle (Read)
        //        bit 3  - ap_ready (Read/COR)
        //        bit 7  - auto_restart (Read/Write)
        //        bit 9  - interrupt (Read)
        //        others - reserved
        // 0x04 : Global Interrupt Enable Register
        //        bit 0  - Global Interrupt Enable (Read/Write)
        //        others - reserved
        // 0x08 : IP Interrupt Enable Register (Read/Write)
        //        bit 0 - enable ap_done interrupt (Read/Write)
        //        bit 1 - enable ap_ready interrupt (Read/Write)
        //        others - reserved
        // 0x0c : IP Interrupt Status Register (Read/COR)
        //        bit 0 - ap_done (Read/COR)
        //        bit 1 - ap_ready (Read/COR)
        //        others - reserved
        // 0x10 : Data signal of iq
        //        bit 31~0 - iq[31:0] (Read/Write)
        // 0x14 : Data signal of iq
        //        bit 31~0 - iq[63:32] (Read/Write)
        // 0x18 : reserved
        // 0x1c : Data signal of event_count
        //        bit 15~0 - event_count[15:0] (Read)
        //        others   - reserved
        // 0x20 : Control signal of event_count
        //        bit 0  - event_count_ap_vld (Read/COR)
        //        others - reserved
        // 0x2c : Data signal of max_events
        //        bit 15~0 - max_events[15:0] (Read/Write)
        //        others   - reserved
        // 0x30 : reserved
        """
        super().__init__(description=description)
        self._buf = None

    def capture(self, max_events=None, wait=True):
        if max_events is None:
            max_events = self.POSTAGE_BUFFER_LEN
        if not self.register_map.CTRL.AP_IDLE:
            getLogger(__name__).debug('Core already capturing, returning existing buffer')
            return self._buf

        # Enable the interrupt
        self.register_map.IP_IER.CHAN0_INT_EN = 1
        self.register_map.GIER = 1
        self.read(0x0C)  # clear it
        self.write(0x2c, min(max(max_events, 1), self.POSTAGE_BUFFER_LEN))
        if self._buf is not None:
            getLogger(__name__).debug('Zeroing and reusing existing buffer')
            assert self._buf.shape == (self.POSTAGE_BUFFER_LEN, self.N_CAPDATA + 1, 2)
            assert self._buf.dtype == np.int16
            self._buf[:] = 0
        else:
            self._buf = buffer.allocate((self.POSTAGE_BUFFER_LEN, self.N_CAPDATA + 1, 2), dtype=np.int16)

        self._buf[:, 0, 0] = -1  # set a sentinal value
        self.write(0x10, np.asarray([self._buf.device_address]).tobytes())
        self.register_map.CTRL.AP_START = 1

        if wait:
            _, int_event = ThreadedPLInterruptManager.get_monitor(self)
            int_event.wait()
            ThreadedPLInterruptManager.remove_monitor(int_event)
        return self._buf

    def get_postage(self, complex=False, scaled=True, rawbuffer=False, omit_first=True):
        """
        Raw takes precedence

        Returns array of event resonator channels and array of [n_events,N_CAPDATA]

        get channels with events via set(ids)
        select events from a channel via events[ids==<your_channel>]
        """
        count = self.event_count
        start = int(omit_first)  # first photon is generally garbage
        if rawbuffer:
            return np.array(self._buf[start:count])
        else:
            return self.process_postage(self._buf[start:count], scaled=scaled, complex=complex)

    @staticmethod
    def process_postage(buffer, complex=True, scaled=True):
        """
        convert postage stamp capture data into rids and associated complex values,
        by default omit the first, garbage stamp
        """
        from ..funcs import postage_buffer_to_data
        return postage_buffer_to_data(buffer, complex=complex, scaled=scaled)

    @property
    def event_count(self):
        return self.read(0x1c)

    def configure(self, max_events=None):
        """ monitor_channels shall be 8 integers in [0,2047] """
        return self.capture(max_events)


class PhotonMAXI(DefaultIP):
    N_PHOTON_BUFFERS = 2  # Must match HLS C
    PHOTON_BUFF_N = 102400  # Must match HLS C and include padding for the partial burst
    PHOTON_DTYPE = _PHOTON_DTYPE
    PHOTON_PACKED_DTYPE = np.uint64
    bindto = ['mazinlab:mkidgen3:photon_maxi:0.2']

    """
    The PhotonMAXI core operates in AP_CONTINUE mode, ending and auto-restarting each time it needs to rotate the buffer
    
    The core reads inbound photons and when it has enough for a maximal (512) burst it writes them to the buffer.
    every time it reads a photon it sets the number of photons written to that buffer (so this number will increment in
    steps of 512 and will generally report that its value is valid while the core is running. 
    
    It appears that during a burst the core continues to ingest photons.
    
    Once a photon is received that has a timestamp with high bits that do not equal the preceding photon OR the number 
    of photons written is at least the rotation quantity the core will burst out one final, partial burst, update the 
    count of photons in that buffer and restart using the next buffer.
    
    The `active_buffer` is the buffer presently being updated. So if active 
    """

    @staticmethod
    def pack_photons(x, out=None):
        ret = np.zeros(x.size, dtype=PhotonMAXI.PHOTON_PACKED_DTYPE) if out is None else out
        ret[:] = (((x['time'] << 12) | x['id']) << 16) | x['phase'].astype(np.uint16)
        return ret

    @staticmethod
    def gen_fake_buffer_data(n, time_us):
        n = int(n)
        data = np.zeros(n, dtype=PhotonMAXI.PHOTON_DTYPE)
        data['time'] = np.sort(np.random.randint(0, high=time_us, size=n))
        data['id'] = np.random.randint(0, high=2047, size=n)
        data['phase'] = np.random.randint(-0x7fff, high=0x7fff, size=n)
        return data

    def __init__(self, description):
        """
        The core watches for trigger events on 8 resonator channels and forwards IQ snippets around trigger events on
        for capture by other parts of the firmware. It is configured with 8 resonator channel values for monitoring.

        // 0x00 : Control signals
        //        bit 0  - ap_start (Read/Write/COH)
        //        bit 1  - ap_done (Read/COR)
        //        bit 2  - ap_idle (Read)
        //        bit 3  - ap_ready (Read/COR)
        //        bit 7  - auto_restart (Read/Write)
        //        bit 9  - interrupt (Read)
        //        others - reserved
        // 0x04 : Global Interrupt Enable Register
        //        bit 0  - Global Interrupt Enable (Read/Write)
        //        others - reserved
        // 0x08 : IP Interrupt Enable Register (Read/Write)
        //        bit 0 - enable ap_done interrupt (Read/Write)
        //        bit 1 - enable ap_ready interrupt (Read/Write)
        //        others - reserved
        // 0x0c : IP Interrupt Status Register (Read/COR)
        //        bit 0 - ap_done (Read/COR)
        //        bit 1 - ap_ready (Read/COR)
        //        others - reserved
        // 0x10 : Data signal of photons_out
        //        bit 31~0 - photons_out[31:0] (Read/Write)
        // 0x14 : Data signal of photons_out
        //        bit 31~0 - photons_out[63:32] (Read/Write)
        // 0x18 : reserved
        // 0x28 : Data signal of active_buffer
        //        bit 7~0 - active_buffer[7:0] (Read)
        //        others  - reserved
        // 0x2c : Control signal of active_buffer
        //        bit 0  - active_buffer_ap_vld (Read/COR)
        //        others - reserved
        // 0x38 : Data signal of photons_per_buf
        //        bit 16~0 - photons_per_buf[16:0] (Read/Write)
        //        others   - reserved
        // 0x3c : reserved
        // 0x40 : Data signal of time_shift
        //        bit 4~0 - time_shift[4:0] (Read/Write)
        //        others  - reserved
        // 0x44 : reserved
        // 0x20 ~
        // 0x27 : Memory 'n_photons' (2 * 17b)
        //        Word n : bit [16:0] - n_photons[n]
        //                 others     - reserved
        """
        super().__init__(description=description)
        self._buf = None
        self._fountain_thread = None

    def get_photons(self, no_copy=False, cautious=True):
        """v0.4 uses 9c7ad901ef20bb356010109c83311bc5e49db81e of photon_maxi
        The active buffer is the index of the buffer being loaded and so the other index should be used
        the active buffer is written every time a photon comes in and so will very likely be valid
        this function takes [0.35, 0.45) ms with cautious=True and ~0.24 um cautious=False
        """
        if cautious and not self.read(0x2c):
            getLogger(__name__).warning(f'active_buffer_ctrl not valid and getting, {repr(self.register_map)}')
            return None

        getLogger(__name__).debug(f'Fetching buffer occupancy and active')
        cnta, cntb, ab = self.mmio.array[8:11]  # about 66% faster that a pair of reads, about 64-140 us
        ab &= 0xff
        count = (cnta if ab else cntb) & 0x1ffff
        getLogger(__name__).debug(f'Slicing data')
        ret = self._buf[0 if ab else 1, :count]

        if not no_copy:
            getLogger(__name__).debug(f'Copy data')
            ret = np.array(ret)

        if cautious and self.read(0x0c):  # The core restarts when buffers are rotated, setting the interrupt
            getLogger(__name__).error(f'Buffers rotated during read. Dropping {count} photons.')
            raise RuntimeError('Buffer changed during read')

        return ret

    def stop_capture(self):
        self.register_map.CTRL.AUTO_RESTART = 0

    def photon_fountain(self, q: (zmq.Socket, queue.SimpleQueue), kill=None, copy_buffer=False, spawn=True,
                        discard_initial=3):
        """
        Is both the target for and can spawn a thread of the same

        q may be a zma socket or a queue.Queue

        kill must be a threading.Event if spawn is not True

        returns an unstarted thread and a kill event to kill the thread

        photons will be fetched and sent over the q copying the buffer if copy_buffer is True

        thread MUST be started before .capture() is called to initiate capture

        """
        if spawn:
            if self._fountain_thread and self._fountain_thread.is_alive():
                raise RuntimeError('Photon fountain already running')
            kill = threading.Event()
            self._fountain_thread = threading.Thread(target=self.photon_fountain, name='photon fountain',
                                                     args=(q,),
                                                     kwargs=dict(kill=kill, spawn=False, copy_buffer=copy_buffer,
                                                                 discard_initial=discard_initial))
            return self._fountain_thread, kill

        assert isinstance(kill, threading.Event)

        log = getLogger(__name__)
        sender = q.put if hasattr(q, 'put') else lambda x: q.send(x, copy=False)

        _, int_event = ThreadedPLInterruptManager.get_monitor(self._interrupts['interrupt']['fullpath'])

        while not kill.is_set():

            int_event.wait(1.1534336)  # 110% max buffer interval, if no interrupt something may be odd so check kill

            if not int_event.is_set():
                log.warning('Timeout waiting for interrupt, is capture started?')
                continue

            int_event.clear()
            self.read(0x0c)

            if discard_initial > 0:
                log.debug(f'Discarding {discard_initial} before sending')
                discard_initial -= 1
                continue

            x = self.get_photons(no_copy=not copy_buffer, cautious=False)

            if int_event.is_set():  # or self.read(0x0c) then we have taken too long
                log.error(f'Buffer likely rotated during fetch, this should not happen.')
                continue

            sender(x)

        ThreadedPLInterruptManager.remove_monitor(int_event)
        sender(b'')

    @property
    def buffer_count_interval(self):
        return self.read(0x38) & 0x1ffff

    @property
    def buffer_interval(self):
        """ Max buffer rotation interval in ms """
        return 2 ** (self.read(0x40) & 0x1f) / 1e3

    @buffer_interval.setter
    def buffer_interval(self, interval_ms):
        """Permissible range is about 0.5 - 1000 ms and will be clipped"""
        l2_buffer_shift = np.round(np.log2(interval_ms * 1000))
        _buffer_time_ms = 2 ** l2_buffer_shift / 1000
        if l2_buffer_shift < 9 or l2_buffer_shift > 20:
            l2_buffer_shift = min(max(l2_buffer_shift, 9), 20)
            getLogger(__name__).warning(f'Requested rotation interval ({interval_ms:.2f} ms) unsupported, '
                                        f'using {2 ** l2_buffer_shift / 1000:.2f}')
        elif abs(_buffer_time_ms - interval_ms) >= .01:
            getLogger(__name__).info(f'Requested rotation interval {interval_ms:.2f} ms rounded to '
                                     f'{_buffer_time_ms:.2f} ms')
        else:
            getLogger(__name__).info(f'Setting rotation interval to {_buffer_time_ms:.2f} ms')

        self.write(0x40, int(l2_buffer_shift))

    def capture(self, n_photons_per_buffer=101888, buffer_time_ms=4.096, cacheable_buffer=True):
        """
        Initiate capture of photons. Any existing buffer will be freed.

        Note that proper functionality of the core requires that at least two photons arrive at least buffer_time_ms
        apart or more than n_photons_per_buffer are seen. If neither of those conditions is met the core will not
        complete. If fewer than 512 photons are seen AND they all arrive within one buffer interval no data will even
        be written into the buffer. Both of these scenarios are essentially impossible and can be resolved by setting
        an arbitrarily low event trigger threshold in the trigger core.

        The FPGA photon buffer is used as a C array of size uint128_t[N_PHOTON_BUFFERS][FLAT_PHOTON_BUFSIZE/2]
        FLAT_PHOTON_BUFSIZE 102400
        N_PHOTON_BUFFERS 2

        Photons are packed in uint64_t and paired for writing, hence the half size axis and 128b word length for
        the core. In the core:
        0,0 [photon0 photon1 buf1] 0,1 [photon2 photon3 buf1] .....
        1,0 [photon0 photon1 buf2] 1,1 [photon2 photon3 buf2] .....

        In python we want to allocate  (N_PHOTON_BUFFERS, FLAT_PHOTON_BUFSIZE)

        the photon count buffer is 17 bits and maxes out at 131071

        the core only increments the count in units of 512 however (a maximal burst of 4KiB of photons)
        then when a threshold is crossed (n_photons_per_buffer) or the high bits of two successive photons
        aren't the same we stop bursting, send off any remaining values, and swap buffers

        the active buffer (preburst) and photon count (post burst) are set once per burst, then the count updated with the
        numbers from the final partial burst (just before restart, buffer swap and zeroing of new buffer count)

        So each buffer must be n_photons_per_buffer + 256*2 photons deep


        Args:
            n_photons_per_buffer: Rotate the buffer approximately every this many photons, underset this by 512 if the
            actual rate is critical. Will be limited to about 100k. FLAT_PHOTON_BUFSIZE-512 (see HLS C code).
            buffer_time_ms: Rotate the buffer approximately every this many ms. Valid range is ~ 0.5 - 1000 us.

        Returns: The buffer. Don't write to this. Don't free this. Glances at the returned buffer should see new
        photons appear in batches of 512 and the buffer count increment in steps of 512, both until a final batch of up
        to 512. Then the other buffer should increment while the values in the inactive buffer stay the same.

        """
        if not self.register_map.CTRL.AP_IDLE:
            getLogger(__name__).debug('Core already capturing, returning existing buffer')
            return self._buf

        if self._buf is not None:
            x = self._buf
            self._buf = None
            x.freebuffer()

        self.register_map.IP_IER.CHAN0_INT_EN = 1  # enable the done interrupt
        self.register_map.GIER = 1

        self.read(0x0C)  # clear any existing interrupt (COR)
        self.read(0x28)  # clear the occupancy valid flag by reading the buffer occupancy (COR)

        self._buf = buffer.allocate((self.N_PHOTON_BUFFERS, self.PHOTON_BUFF_N),
                                    dtype=self.PHOTON_PACKED_DTYPE, cacheable=cacheable_buffer)

        self.write(0x10, np.asarray([self._buf.device_address]).tobytes())

        self.buffer_interval = buffer_time_ms

        n_photons_per_buffer = int(n_photons_per_buffer)
        if n_photons_per_buffer > self.PHOTON_BUFF_N - 512:
            getLogger(__name__).warning(f'Requesting too large a buffer, buffer will be rotated every '
                                        f'={self.PHOTON_BUFF_N} photons')

        self.write(0x38, min(max(int(n_photons_per_buffer), 1), self.PHOTON_BUFF_N - 512))
        self.register_map.CTRL.AUTO_RESTART = 1
        self.register_map.CTRL.AP_START = 1
        return self._buf

    def configure(self):
        """ monitor_channels shall be 8 integers in [0,2047] """
        return self.capture()


class PhotonPacketizer(DefaultIP):
    MAX_PHOTON_PACKET_N = 262144  # Must not exceed dma buffer size
    PHOTON_DTYPE = _PHOTON_DTYPE
    PHOTON_PACKED_DTYPE = np.uint64
    bindto = ['mazinlab:mkidgen3:photon_fifo_packetizer:0.1']

    @staticmethod
    def pack_photons(x, out=None):
        ret = np.zeros(x.size, dtype=PhotonMAXI.PHOTON_PACKED_DTYPE) if out is None else out
        ret[:] = (((x['time'] << 12) | x['id']) << 16) | x['phase'].astype(np.uint16)
        return ret

    @staticmethod
    def unpack_photons(x, out=None, n=0):
        """
        Unpack packed photons, optionally accumulating them into an existing output array

        Args:
            x: an array of packed photons
            out: optional, an array of type PhotonMAXI.PHOTON_DTYPE with the shape of x, but see n
            n: optional, an index into out to insert unpacked photons, the first axis is used if >1d an
                IndexError is raised if there is insufficient space.

        Returns: the unpacked photon array

        """
        if out is None:
            n = 0
        elif x.shape[0] + n > out.shape[0]:
            raise IndexError('Output array is too small')

        ret = np.zeros(x.shape, dtype=PhotonMAXI.PHOTON_DTYPE) if out is None else out
        sl = slice(n, n + x.shape[0])
        ret['phase'][sl] = x & 0xffff
        ret['time'][sl] = x >> 28
        ret['id'][sl] = (x >> 16) & 0xfff
        return ret

    @staticmethod
    def gen_fake_buffer_data(n, time_us):
        n = int(n)
        data = np.zeros(n, dtype=PhotonMAXI.PHOTON_DTYPE)
        data['time'] = np.sort(np.random.randint(0, high=time_us, size=n))
        data['id'] = np.random.randint(0, high=2047, size=n)
        data['phase'] = np.random.randint(-0x7fff, high=0x7fff, size=n)
        return data

    def __init__(self, description):
        """
        The core watches for trigger events on 8 resonator channels and forwards IQ snippets around trigger events on
        for capture by other parts of the firmware. It is configured with 8 resonator channel values for monitoring.

        // 0x10 : Data signal of photons_per_buf
        //        bit 16~0 - photons_per_buf[16:0] (Read/Write)
        //        others   - reserved
        // 0x18 : Data signal of time_shift
        //        bit 4~0 - time_shift[4:0] (Read/Write)
        //        others  - reserved
        """
        super().__init__(description=description)

    @property
    def buffer_count_interval(self):
        return self.read(0x10) & 0x1ffff

    @property
    def buffer_interval(self):
        return 2 ** (self.read(0x18) & 0x1f) / 1e3

    @buffer_interval.setter
    def buffer_interval(self, interval_ms):
        """Permissible range is about 0.5 - 1000 ms and will be clipped"""
        l2_buffer_shift = np.round(np.log2(interval_ms * 1000))
        _buffer_time_ms = 2 ** l2_buffer_shift / 1000
        if l2_buffer_shift < 9 or l2_buffer_shift > 20:
            l2_buffer_shift = min(max(l2_buffer_shift, 9), 20)
            getLogger(__name__).warning(f'Requested rotation interval ({interval_ms:.2f} ms) unsupported, '
                                        f'using {2 ** l2_buffer_shift / 1000:.2f}')
        elif abs(_buffer_time_ms - interval_ms) >= .01:
            getLogger(__name__).info(f'Requested rotation interval {interval_ms:.2f} ms rounded to '
                                     f'{_buffer_time_ms:.2f} ms')
        else:
            getLogger(__name__).info(f'Setting rotation interval to {_buffer_time_ms:.2f} ms')

        self.write(0x18, int(l2_buffer_shift))

    def packetize(self, n_photons_per_buffer=100000, buffer_time_ms=4.096):
        """
        Initiate packetization of photons.

        Note that proper functionality of the core requires that at least two photons arrive at least buffer_time_ms
        apart or more than n_photons_per_buffer are seen. If neither of those conditions is met the core will not
        complete. If fewer than 512 photons are seen AND they all arrive within one buffer interval no data will even
        be written into the buffer. Both of these scenarios are essentially impossible and can be resolved by setting
        an arbitrarily low event trigger threshold in the trigger core.

        Args:
            n_photons_per_buffer: Rotate the buffer approximately every this many photons, underset this by 512 if the
            actual rate is critical. Will be limited to about 100k. FLAT_PHOTON_BUFSIZE-512 (see HLS C code).
            buffer_time_ms: Rotate the buffer approximately every this many ms. Valid range is ~ 0.5 - 1000 us.

        """
        self.buffer_interval = buffer_time_ms

        n_photons_per_buffer = int(n_photons_per_buffer)
        if n_photons_per_buffer > self.MAX_PHOTON_PACKET_N:
            getLogger(__name__).warning(f'Requesting too large a packet, using packet of '
                                        f'{self.MAX_PHOTON_PACKET_N} photons.')

        self.write(0x10, min(max(n_photons_per_buffer - 2, 1), self.MAX_PHOTON_PACKET_N - 2))

    def configure(self):
        """ monitor_channels shall be 8 integers in [0,2047] """
        return self.packetize()



from mkidgen3.funcs import *
import logging
import numpy as np
import numpy.typing as nt
import platform
import matplotlib.pyplot as plt

from mkidgen3.system_parameters import (ADC_DAC_INTERFACE_WORD_LENGTH, DAC_RESOLUTION, DAC_SAMPLE_RATE, SYSTEM_BANDWIDTH,
                                        DAC_FREQ_RES, DAC_FREQ_MAX, DAC_FREQ_MIN)


def _same(a, b):
    """quick test for array vs None"""
    if not isinstance(a, type(b)):
        return False
    if a is None and b is None:
        return True
    try:
        return np.all(a == b)
    except:
        pass
    try:
        return a == b
    except:
        return False


class Waveform:
    @property
    def output_waveform(self):
        """Subclasses shall implement _values """
        return self._values

    @property
    def sample_rate(self):
        """Subclasses shall implement _sample_rate """
        return self._sample_rate


class TabulatedWaveform(Waveform):
    """
    Use this class if you want to pass existing tabulated values directly to the DAC LUT without any scaling or
    computation
    """

    def __init__(self, tabulated_values=None, sample_rate=DAC_SAMPLE_RATE):
        self._values = tabulated_values
        self._sample_rate = sample_rate

    def __str__(self):
        return f'TabulatedWaveform with sample rate {self._sample_rate}'

    def __ne__(self, other):
        return not (self._sample_rate == other.sample_rate and
                    _same(self._values, other._values))


class SimpleFreqlistWaveform(Waveform):
    def __init__(
            self,
            frequencies,
            amplitudes=None,
            phases=None,
            n_samples=(1 << 19),
            sample_rate=4.096e9,
            allow_sat=False,
    ):
        self.freqs = quantize_frequencies(
            np.asarray(frequencies), rate=sample_rate, n_samples=n_samples
        )
        if amplitudes is None:
            self.amps = np.ones_like(self.freqs) / self.freqs.size
        else:
            self.amps = np.asarray(amplitudes)
        if phases is None:
            self.phases = np.random.random(self.freqs.size)*2*np.pi
        else:
            self.phases = np.asarray(phases)
        self.n_samples = n_samples
        self._sample_rate = sample_rate
        self.allow_sat = allow_sat

    @property
    def quant_freqs(self):
        return self.freqs

    @property
    def _values(self):
        fft_freqs = np.fft.fftfreq(self.n_samples, 1 / self.sample_rate)
        fft = np.zeros_like(fft_freqs, dtype=np.complex128)
        for i, f in enumerate(self.freqs):
            idx = np.argmin(np.abs(fft_freqs - f))
            fft[idx] += self.amps[i] * np.exp(1.j * self.phases[i]) * self.n_samples
        data = np.fft.ifft(fft)

        if self.allow_sat:
            data.real = np.clip(data.real, -1, 1)
            data.imag = np.clip(data.imag, -1, 1)
        else:
            if np.max(np.abs(data.real)) > 1 or np.max(np.abs(data.imag)) > 1:
                raise RuntimeError(
                    "Data exceeded DAC output range i: ({:f}, {:f}), q: ({:f}, {:f})".format(
                        np.min(data.real),
                        np.max(data.real),
                        np.min(data.imag),
                        np.max(data.imag),
                    )
                )
        return data * ((1 << 15) - 6)


class FreqlistWaveform(Waveform):
    def __init__(self, frequencies: Iterable[int | float] = None, n_samples: int = DAC_LUT_SIZE,
                 sample_rate: int | float = DAC_SAMPLE_RATE, amplitudes: Iterable[int | float] = None,
                 phases: Iterable[int | float] = None, iq_ratios: Iterable[int | float] = None,
                 phase_offsets: Iterable[int | float] = None, seed: int = 2, dac_dynamic_range: float = 1.0,
                 optimize_phase: bool = True, compute: bool = False):
        """
        Args:
            frequencies: frequencies in the waveform [Hz]
            n_samples: number of complex samples in the waveform
            sample_rate: waveform sample rate [Hz] (should be the DAC sample rate)
            amplitudes: amplitudes of each tone in the waveform. If None, all ones is assumed
            phases: phases of each tone in the waveform in [0, 2*pi). If None, random phases are generated using seed
            iq_ratios: ratios for IQ values used to help minimize image tones in band. Allowed values between 0 and 1
                       If None, 50:50 ratio (all ones) is assumed
            phase_offsets: phase offsets in [0, 2*np.pi)
            seed: random seed to seed phase randomization process
            dac_dynamic_range: how much of dac dynamic range to use. Allow values [0.0,1.0]. Default is 1.0 (all)
            optimize_phase: check quantization error and re-generate waveform with new random phases if too large
            compute: compute waveform
        """
        self.freqs = np.asarray(frequencies)
        assert (DAC_FREQ_MIN <= self.freqs).all() and (self.freqs <= DAC_FREQ_MAX).all(), (f"freqs must be in "
                                                                                      f"[{DAC_FREQ_MIN}, "
                                                                                      f"{DAC_FREQ_MAX}]")
        self.n_samples = n_samples
        self.amps = amplitudes if amplitudes is not None else np.ones_like(frequencies)
        self._sample_rate = sample_rate
        self._optimize_phase = optimize_phase

        if phases is None:
            self.phases = np.random.default_rng(seed=seed).uniform(0, 2 * np.pi, size=self.freqs.size)
        else:
            self.phases = np.asarray(phases)
        assert (0 <= self.phases).all and (self.phases < 2*np.pi).all, "phases must be between 0 and 2 pi"
        self.dac_dynamic_range = dac_dynamic_range

        self.iq_ratios = np.asarray(iq_ratios) if iq_ratios is not None else np.ones_like(frequencies)
        self.phase_offsets = np.asarray(phase_offsets) if phase_offsets is not None else np.zeros_like(frequencies)
        assert (0 <= self.phase_offsets).all and (self.phase_offsets < 2*np.pi).all, ("phase offsets must be between 0 "
                                                                                      "and 2 pi")
        self.quant_freqs = quantize_frequencies(self.freqs, rate=sample_rate, n_samples=n_samples)

        self._seed = seed
        self.__values = None  # cache

        self.quant_vals = None
        self.quant_error = None

        if compute:
            self.output_waveform

    def __ne__(self, other):
        if self.quant_vals is not None and other.n_samples is not None:
            return not _same(self.quant_vals, other.quant_vals)

        return not (self.dac_dynamic_range == other.dac_dynamic_range and
                    _same(self.iq_ratios, other.iq_ratios) and
                    self.n_samples == other.n_samples and
                    _same(self.phases, other.phases) and
                    _same(self.quant_freqs, other.quant_freqs) and
                    self._sample_rate == other._sample_rate and
                    _same(self.amps, other.amps) and
                    _same(self.phase_offsets, other.phase_offsets))

    def __repr__(self):
        return f'<{str(self)}>'

    def __str__(self):
        preview_dict = {'freqs': self.freqs, 'amps': self.amps, 'phases': self.phases,
                        'iq_ratios': self.iq_ratios, 'phase_offsets': self.phase_offsets,
                        'quant_error': self.quant_error}
        for key, value in preview_dict.items():
            if value is None or (value.size < 3):
                preview_dict[key] = value
            else:
                preview_dict[key] = value[:3]

        return f'FreqlistWaveform: {preview_dict}'

    @property
    def _values(self) -> nt.NDArray[np.complex128]:
        """
        Return or calculate waveform values
        Returns: Complex values where the real and imag part have been quantized to ints in accordance with specified
                 dac dynamic range

        """
        if self.quant_vals is None:
            self.__values = self._compute_waveform()

            self.quant_vals, self.quant_error = quantize_to_int(self.__values, resolution=DAC_RESOLUTION, signed=True,
                                                                word_length=ADC_DAC_INTERFACE_WORD_LENGTH,
                                                                dyn_range=self.dac_dynamic_range, return_error=True)
            if self._optimize_phase:
                self._optimize_random_phase(max_quant_err=1 * predict_quantization_error(resolution=DAC_RESOLUTION),
                                            max_attempts=3)
        return self.quant_vals

    def _compute_waveform(self, phases: Iterable | None = None) -> nt.NDArray[np.complex64]:
        """
        Compute the raw waveform with no scaling or casting.
        Args:
            phases: new phases to compute waveform with (useful for re-generating random phases)

        Returns: Raw waveform values.

        """
        iq = np.zeros(self.n_samples, dtype=np.complex64)
        # generate each signal
        t = 2 * np.pi * np.arange(iq.size) / self._sample_rate

        phases = self.phases if phases is None else phases

        if self.phase_offsets.any() or (self.iq_ratios != 1).any() or self.phases.ndim > 1:
            logging.getLogger(__name__).debug(
                f'Computing net waveform with {self.freqs.size} tones in a loop to apply IQ ratios and phase offsets.\n'
                f'For 2048 tones this takes about 7 min.')
            for i in range(self.freqs.size):
                exp = self.amps[i] * np.exp(1j * (t * self.quant_freqs[i] + phases[i]))
                scaled = np.sqrt(2) / np.sqrt(1 + self.iq_ratios[i] ** 2)
                c1 = self.iq_ratios[i] * scaled * np.exp(1j * np.deg2rad(self.phase_offsets)[i])
                iq.real += c1.real * exp.real + c1.imag * exp.imag
                iq.imag += scaled * exp.imag

        else:
            logging.getLogger(__name__).debug(
                f'Computing net waveform with {self.freqs.size} tones using IFFT.')
            possible_tones = np.linspace(-DAC_SAMPLE_RATE/2, (DAC_SAMPLE_RATE/2)-DAC_FREQ_RES, DAC_LUT_SIZE)
            tone_idxs = np.concatenate([np.where(possible_tones == freq) for freq in self.quant_freqs]).flatten()
            fft = np.zeros(2**19, dtype=np.complex64)
            for tone_number, tone_idx in enumerate(tone_idxs):
                fft[tone_idx] = self.amps[tone_number]*np.exp(1j*self.phases[tone_number])
            iq = np.fft.ifft(np.fft.fftshift(fft))

        return iq

    def _optimize_random_phase(self,
                               max_quant_err: float | int = 1 * predict_quantization_error(resolution=DAC_RESOLUTION),
                               max_attempts: int = 3) -> None:
        """
        Regenerate random phases, waveform values, and quantized values if quantization error is too high.
        Args:
            max_quant_err: maximum absolute allowable quantization error defined as abs(expected value - achieved value)
            max_attempts: maximum numer of times to regenerate random phases, waveform values, and quantized values

        Returns: None

        Waveform random phases, waveform values, quantized values, and quantization error are only updated if a
        solution is found.
        """

        if max_quant_err is None:
            max_quant_err = 3 * predict_quantization_error(resolution=DAC_RESOLUTION)

        if self.quant_error < max_quant_err:  # already optimal
            return

        quant_error = self.quant_error
        cnt = 0
        while quant_error > max_quant_err:
            logging.getLogger(__name__).warning(
                f"Quantization error {quant_error} exceeded max quantization error {max_quant_err}. The freq comb's "
                f"relative phases may have added up sub-optimally. Calculating with new random phases")
            self._seed += 1
            self.phases = np.random.default_rng(seed=self._seed).uniform(0., 2. * np.pi, len(self.freqs))
            values = self._compute_waveform(phases=self.phases)
            quant_vals, quant_error = quantize_to_int(values, resolution=DAC_RESOLUTION, signed=True,
                                                      word_length=ADC_DAC_INTERFACE_WORD_LENGTH,
                                                      return_error=True)
            cnt += 1
            if cnt > max_attempts:
                raise Exception("Process reach maximum attempts: Could not find solution below max quantization error.")

        self.__values = values
        self.quant_vals, self.quant_error = quant_vals, quant_error


def WaveformFactory(n_uniform_tones=None, output_waveform=None, frequencies=None,
                    n_samples=DAC_LUT_SIZE, sample_rate=DAC_SAMPLE_RATE, amplitudes=None, phases=None,
                    iq_ratios=None, phase_offsets=None, seed=2, dac_dynamic_range=1.0, compute=False):
    if output_waveform is not None:
        return TabulatedWaveform(tabulated_values=output_waveform, sample_rate=sample_rate)
    if n_uniform_tones is not None:
        if n_uniform_tones not in (512, 1024, 2048):
            raise ValueError('Requested number of power sweep tones not supported. Allowed values are 512, 1024, 2048.')
        frequencies = uniform_freqs(n_uniform_tones, bandwidth=SYSTEM_BANDWIDTH)
    if frequencies is None:
        return None
    frequencies = np.asarray(frequencies)
    return FreqlistWaveform(frequencies=frequencies, n_samples=n_samples, sample_rate=sample_rate,
                            amplitudes=amplitudes, phases=phases, iq_ratios=iq_ratios, phase_offsets=phase_offsets,
                            seed=seed, dac_dynamic_range=dac_dynamic_range, compute=compute)


if __name__ == '__main__':
    """
    Test code for debugging waveform generation.
    """
    from mkidgen3.util import pseudo_random_tones
    from mkidgen3.server.feedline_config import *
    import matplotlib.pyplot as plt

    tone = 400e6
    exclude = np.array([tone - 4e6, tone - 3e6, tone - 2e6, tone - 1e6, tone, tone + 1e6, tone + 2e6, tone + 3e6, tone + 4e6])
    random_tones = pseudo_random_tones(n=2048, buffer=300e3, spread=True, exclude=exclude)
    wvfm_tones = np.append(np.array([tone]), random_tones)
    wvfm_cfg = WaveformConfig(
        waveform=WaveformFactory(frequencies=wvfm_tones, seed=5, dac_dynamic_range=1 / 150, compute=True))
    wvfm_fft = np.abs(np.fft.fftshift(np.fft.fft(wvfm_cfg.waveform.output_waveform)))
    freqs = np.linspace(-2.048e9, 2.048e9 - 7.8125e3, 2 ** 19)
    plt.plot(freqs, wvfm_fft)
    plt.xlim(395e6, 405e6)
    plt.show()
    print('hi')



import numpy as np
from fpbinary import FpBinary, OverflowEnum, RoundingEnum

FP16_15 = lambda x: FpBinary(int_bits=1, frac_bits=15, signed=True, value=x)
FP16_15 = lambda x: FpBinary(int_bits=1, frac_bits=15, signed=True, value=x)
FP26_26 = lambda x: FpBinary(int_bits=0, frac_bits=26, signed=True, value=x)
FP16_26 = lambda x: FpBinary(int_bits=-10, frac_bits=26, signed=True, value=x)
FP32_31 = lambda x: FpBinary(int_bits=1, frac_bits=31, signed=True, value=x)
FP32_8 = lambda x: FpBinary(int_bits=32 - 9, frac_bits=8, signed=True, value=x)
FP18_17 = lambda x: FpBinary(int_bits=1, frac_bits=17, signed=True, value=x)
FP16_25 = lambda x: FpBinary(int_bits=-9, frac_bits=25, signed=True, value=x)
FP18_25 = lambda x: FpBinary(int_bits=-7, frac_bits=25, signed=True, value=x)
FP16_17 = lambda x: FpBinary(int_bits=-1, frac_bits=17, signed=True, value=x)


def fp_factory(int, frac, signed, frombits=False, include_index=False):
    if isinstance(signed, str):
        signed = True if 'signed' == signed.lower() else False
    else:
        signed = bool(signed)
    if frombits:
        return lambda x: FpBinary(int_bits=int, frac_bits=frac, signed=signed, bit_field=x)
    else:
        if include_index:
            return lambda x: FpBinary(int_bits=int, frac_bits=frac, signed=signed, value=x).__index__()
        else:
            return lambda x: FpBinary(int_bits=int, frac_bits=frac, signed=signed, value=x)


def fparray(a, fpgen=None):
    """
    Convert a numpy array to a numpy object array of FpBinary numbers.

    Complex numbers gain an additional axis of length 2.

    if fpgen is None 16wide  1integer signed is used
    """
    try:
        fpgen = lambda x: FpBinary(int_bits=fpgen[0], frac_bits=fpgen[1], signed=fpgen[2], value=x)
    except Exception:
        pass
    if fpgen is None:
        fpgen = lambda x: FpBinary(int_bits=1, frac_bits=15, signed=True, value=x)

    if a.dtype in (np.complex, np.complex128, np.complex64):
        out = np.empty(tuple(list(a.shape)+[2]), dtype=object)
        b = [(fpgen(x.real), fpgen(x.imag)) for x in a.flat]
        for i in range(out.size):
            out.flat[i] = b[i//2][i % 2]
        #You'd think this would work, but it doesn't as the FP types get cast to bool
        #out.flat = [(fpgen(x.real), fpgen(x.imag)) for x in a.flat]
    else:
        out = np.empty_like(a, dtype=object)
        b = [fpgen(x) for x in a.flat]
        for i in range(out.size):
            out.flat[i] = b[i]
        #out.flat = [fpgen(x) for x in a.flat]
    return out



from mkidgen3.server.waveform import WaveformFactory, FreqlistWaveform
import numpy as np
from mkidgen3.system_parameters import DAC_SAMPLE_RATE, DAC_FREQ_RES, DAC_LUT_SIZE
import scipy.signal as sig
import pytest

test_wvfms = [WaveformFactory(frequencies=[200e6, 400e6 + 7812.5], amplitudes=[2, 3], compute=True),
              WaveformFactory(n_uniform_tones=2048, compute=True),
              WaveformFactory(n_uniform_tones=2048, dac_dynamic_range=0.8, compute=True)]


@pytest.mark.parametrize("wvfm", test_wvfms)
def test_wvfm_tones(wvfm: FreqlistWaveform) -> None:
    """
    Inspect generated waveform via FFT and ensure the tones are at the right frequencies.
    """
    fft_abs = np.abs(np.fft.fftshift(np.fft.fft(wvfm.quant_vals)))
    tones, _ = sig.find_peaks(fft_abs, height=np.mean(fft_abs))
    possible_tones = np.linspace(-DAC_SAMPLE_RATE / 2, (DAC_SAMPLE_RATE / 2) - DAC_FREQ_RES, DAC_LUT_SIZE)

    assert (possible_tones[tones] == wvfm.quant_freqs).all, "Tones are not in the correct place"


@pytest.mark.parametrize("wvfm", test_wvfms)
def test_wvfm_phases(wvfm: FreqlistWaveform) -> None:
    """
    Inspect generated waveform via FFT and ensure the phases are correct.
    """
    fft = np.fft.fftshift(np.fft.fft(wvfm.quant_vals))
    tones, _ = sig.find_peaks(np.abs(fft), height=np.mean(np.abs(fft)))
    phases = np.angle(fft[tones])
    assert (np.isclose(wvfm.phases, phases)).all, "Phases are wrong"



from .captures import *
from zmq import *
from .pixelmap import PixelMap
from mkidgen3.mkidpynq import unpack_photons

MAX_PHOTON_PER_CR_SEND = 5000
INSTRUMENT_PHOTON_TYPE = (('time', 'u32'), ('x', 'u32'), ('y', 'u32'), ('phase', 'u16'))

class Gen3TableSaver:
    def __init__(self, file, size_hint=None):
        self._file=file

    def grow(self, photons):
        pass


class Gen3LiveImage:
    def __init__(self, destination, exposure_time, photon_rate=True):
        self.exposure_time = exposure_time
        self.photon_rate = photon_rate

    def add_photons(self, photons):
        pass


class PhotonAggregator:
    """ A stub for combining photon capture data from multiple feedlines into a cohesive entity suitable for further
    use. Photon CaptureJobs result in photon data being published by the various feedline readout servers under the
    associated capture ID. To gather data from a full array we need to connect to each stream receive the data,
    uncompress it, convert the timestamps to full unix times, convert the feedline channels to full readout channels (
    FL+FLchannel), and then, depending, do things like convert to a rate, array image, save send on, or so forth

    Captures can die or be killed
    """
    def __init__(self, time_offset, jobs: List[CaptureJob] or None, pixel_map:PixelMap):
        self._time_offset = time_offset
        self._new_jobs = jobs
        self._map = pixel_map
        self._dead_jobs = []
        self._jobs = []
        self._hdf_saver = None

    def add_capture(self, jobs: List[CaptureJob]):
        self._new_jobs.extend(jobs)

    def run(self):
        poller = zmq.Poller()
        poller.register(self._pipe[1], flags=zmq.POLLIN)

        getLogger(__name__).debug(f'Listening for data')
        self._pipe[1].send(b'')

        while True:

            for job in list(self._new_jobs):
                assert not job.datasink.is_alive()
                poller.register(job.datasink.establish())  #TODO this isn't thread safe if the di
                self._jobs.append(job)
                self._new_jobs.remove(job)

            avail = dict(poller.poll())
            if self._pipe[1] in avail:
                getLogger(__name__).debug(f'Received shutdown order, terminating {self}')
                break

            nnew = 0
            photons_buf = np.recarray(len(avail)*MAX_PHOTON_PER_CR_SEND, dtype=INSTRUMENT_PHOTON_TYPE)

            for job in (j for j in list(self._jobs) if j.datasink.socket in avail):
                raw_phot_data = job.datasink.receive()

                if raw_phot_data is None:
                    getLogger(__name__).debug(f'Photon data stream over for {job}')
                    poller.unregister(job.datasing.socket)
                    self._jobs.remove(job)
                    self._dead_jobs.append(job)
                    continue

                sl_out = slice(nnew, nnew+raw_phot_data.size)
                photons_buf['time'][sl_out] = raw_phot_data['time'] + self._time_offset
                photons_buf['phase'][sl_out] = raw_phot_data['phase']
                xy = self._map[job.fl_ndx, raw_phot_data['id']]  #todo
                photons_buf['x'][sl_out] = xy[0]
                photons_buf['y'][sl_out] = xy[1]
                nnew+=raw_phot_data.size

            if self._live_image is not None:
                self._image_image.add_photons(photons_buf[:nnew])

            if self._hdf_saver is not None:
                self._hdf_saver.grow(photons_buf[:nnew])

            if self._republisher:
                self._republisher.send(photons_buf[:nnew])

        for job in self._jobs:
            job.datasink.destablish()


def make_image(map, photons, timestep_us=None, rate=False):
    """ Build an image out of the current photon data, specify a timestep (in us) to make a timecube"""

    first, last = photons['time'][[0, -1]]
    duration_s = (last - first) / 1e6
    if first >= last:
        raise ValueError('Last photon time is not after the first!')

    if timestep_us:
        timebins = (np.arange(first, last + timestep_us, timestep_us, dtype=type(timestep_us)),)
    else:
        timebins = tuple()
    imbins = map.map_bins

    val = map.map[photons['id']]
    val = (val,) if map.map.ndim == 1 else (val[0], val[1])
    coord = val + (photons['time'],) if timestep_us else val

    hist, _ = np.histogramdd(coord, bins=imbins + timebins, density=rate)
    if rate:
        hist *= photons.size
        hist /= duration_s if timestep_us is None else 1e6

    return hist

from mkidgen3.mkidpynq import PHOTON_DTYPE
class PhotonAccumulator:
    def __init__(self, max_ram=100, pixelmap: PixelMap = None, n_channels=None):
        """
        A tool to gather photons together without running out of ram and ease access
        to timeseries, images, and other helpful metrics.

        max_ram: how much space to allow for photon storage
        pixelmap: and optional channel id to x or xy map, should be a nChan x 1 or 2 array positions
        n_channels: optional shortcut if pixelmap not specified
        """
        assert bool(n_channels) ^ bool(pixelmap), 'Must specify n_channels xor pixelmap'
        self.ram_limit = max_ram  # MiB
        self.map = pixelmap if pixelmap else PixelMap(np.arange(n_channels, dtype=int), n_channels)
        self._data = np.zeros(self.ram_limit * 1024 ** 2 // PHOTON_DTYPE.itemsize, dtype=PHOTON_DTYPE)
        self.n = 0
        self.accumulations = 0

    def __getitem__(self, n):
        """Get photons for channel n"""
        return self.data[self.data['id'] == n]

    def accumulate(self, buffer, n=None):
        """Add some packed photons to the pot"""
        self.accumulations += 1
        if n is None:
            d = np.array(buffer)
            n = buffer.size
        else:
            d = np.array(buffer[:n])

        drop = max(n - (self._data.size - self.n), 0)
        if drop:
            getLogger(__name__).debug(f'Memory used up, dropping {drop} old photons.')
            self._data[:self._data.size - drop] = self._data[drop:]
        unpack_photons(buffer[:n], out=self._data, n=self.n)
        self.n += n - drop

    @property
    def data(self):
        return self._data[:self.n]

    def image(self, timestep_us=None, rate=False):
        """ Build an image out of the current photon data, specify a timestep (in us) to make a timecube"""
        return make_image(self.map, self.data[:self.n], timestep_us=timestep_us, rate=rate)



from mkidgen3.server import pixelmap
if __name__=='__main__':

    h5file = 'test.h5'
    live_image = ('live.tcp', 'live.bmp')
    repub = ('photons.tcp')

    #Need to manually spin up feedline servers

    N_FEEDLINES, N_RESONATORS = 2, 2048
    frs1 = FRSClient(url1)
    frs2 = FRSClient(url2)
    cfg = FeedlineConfig()

    servers = [frs1, frs2]
    map = pixelmap.example_map(N_FEEDLINES, N_RESONATORS)
    jobs = [CaptureJob(CaptureRequest(.5, 'photons', cfg, frs),
                       submit=(False, False)) for frs in servers]

    aggregator = PhotonAggregator(int(time.time()), jobs, map)



import numpy as np

#
# class PixelMap:
#     """ A stub for the mapping of feedline and channel to pixel position
#
#     map=PixelMap()
#     map[fl, channel] -> array of xy for specified fl, channel
#     unplaced channels shall be non-finite
#     """
#     def __init__(self, x):
#         """
#         Args:
#             x: an n_feedline x n_channels x 2 array where the last dimension is x and y
#             x[1,300] would then be the x,y of feedline 1 channel 300
#             finite elements must be unique (np.unique(x[np.isfinite(x)]).size==np.isfinite(x).sum())
#         """
#         assert np.unique(x[np.isfinite(x)]).size==np.isfinite(x).sum()
#         self._map = x
#
#     def __getitem__(self, fl, channel):
#         return self._map[fl, channel]


class PixelMap:
    """ The attribute map_bins is a tuple of the pixel bin edges i.e. for histograms"""

    def __init__(self, map, nx, ny=0):
        """
        A Pixel to channel mapping

        nx: the number of x pixels
        ny: the number of y pixels, 0 if 1d
        map: is a 1 or 2d array channel id to pixel coordinate. 1d is for linear arrays or
        single feedlines. Will be coerced to integer.

        """
        self.map = map.astype(int)
        self.nx = max(nx, 1)
        self.ny = max(ny, 0)
        assert (ny > 0) == (map.ndim == 2)
        # TODO assertion that no pixel is mapped to more than one channel
        #  set(zip(self.map[:,0], self.map[:,1]))

        maxn = self.map.max(axis=-1)
        ok = maxn[0] < self.nx and maxn[1] < self.ny if ny else maxn < self.nx
        assert ok, "Mapped pixels out of bounds"

        if not ny:
            self.map_bins = (np.arange(self.nx + 1, dtype=int),)
        else:
            self.map_bins = np.arange(self.nx + 1, dtype=int), np.arange(self.ny + 1, dtype=int)


def example_map(n_feedlines, n_channels)->PixelMap:
    """

    Args:
        n_feedlines: The total number of feed lines present
        n_channels:

    Returns: An example map

    """
    return PixelMap(np.arange(n_feedlines*n_channels, dtype=int).reshape((n_feedlines, n_channels)),
                    n_feedlines, n_channels)



import logging.config
import numpy as np
import os
import yaml
import importlib.resources
from logging import getLogger
import zmq
from mkidgen3.opfb import opfb_bin_number, opfb_bin_center
import numpy.typing as nt
from mkidgen3.system_parameters import (OPFB_CHANNEL_SAMPLE_RATE, ADC_SAMPLE_RATE, N_OPFB_CHANNELS, DAC_FREQ_MIN,
                                        DAC_FREQ_MAX, DAC_FREQ_RES)
from typing import Iterable
import subprocess


def pseudo_random_tones(n: int, buffer: float = 300e3, spread: bool = True, exclude: Iterable = None) -> (
        nt.NDArray)[float]:
    """
    Generate a 1D array of tones where each tone in randomly placed in an OPFB bin
    Args:
        n: number of tones
        buffer: bandwidth from the edges of the bin where tones will not be placed[Hz]
        spread: use every other opfb channels so tones are more spread out.
        exclude: list of tones whose bin will not have a tone placed anywhere in the same bin.
    Returns:
        A 1D array of tones where there's 1 tone per OPFB bin and buffer away from the upper and lower bin edges.
    Tones are placed symmetrically around DC, 1 per bin with the specified buffer

    """
    opfb_halfband = OPFB_CHANNEL_SAMPLE_RATE / 2
    exclude = np.asarray(exclude)
    assert n % 2 == 0, "Only even number of tones is supported."
    assert n < 4095, "Max number of tones is 4095 (one per bin excluding DC bin)."
    assert buffer < opfb_halfband, f"Buffer size is larger than channel width, max buffer allowed is {opfb_halfband}."
    rand_offsets = np.random.default_rng(seed=2).uniform(low=buffer-opfb_halfband, high=opfb_halfband-buffer, size=n)
    bc = (ADC_SAMPLE_RATE / N_OPFB_CHANNELS) * np.linspace(-N_OPFB_CHANNELS / 2, N_OPFB_CHANNELS / 2 - 1,
                                                                    N_OPFB_CHANNELS)
    if spread:
        bc = bc[::2]
    tone_bin_centers = np.concatenate((bc[bc.size//2-n//2:bc.size//2], bc[bc.size//2:+bc.size//2+n//2]))

    if exclude.any():
        exclude_bin_centers = opfb_bin_center(opfb_bin_number(exclude, ssr_raw_order=False), ssr_order=False)
        exclude_idx = []
        for i, tone_bin in enumerate(tone_bin_centers):
            if (tone_bin == exclude_bin_centers).any():
                exclude_idx.append(i)
        tone_bin_centers = np.delete(tone_bin_centers, exclude_idx)
        rand_offsets = np.delete(rand_offsets, exclude_idx)
    return np.clip(tone_bin_centers + rand_offsets, DAC_FREQ_MIN+DAC_FREQ_RES, DAC_FREQ_MAX-DAC_FREQ_RES)



def compute_max_val(x) -> float:
    return max(x.real.max(), x.imag.max(), np.abs(x.imag.min()), np.abs(x.imag.min()))


def rx_power(data: nt.NDArray[np.int32] | nt.NDArray[np.complex64]) -> tuple[float, float, float]:
    """
    Compute ADC average power received in milliwatts and dBm and maximum vppd.
    Args:
        data: ADC capture data: complex and integer data normalized to max val = 1.0 or raw ints are all supported
              normalized non-complex data will pass straight through to the power calculations.

    Returns: tuple: Average Power in mW, Average power in dBm, Maximum VPPD seen by ADC

    """
    if np.iscomplex(data).any():
        if np.any(abs(data.real) > 1.0) or np.any(abs(data.imag) > 1.0):  # normalize
            data = np.concatenate((np.int16(data.real[:, np.newaxis]), np.int16(data.imag[:, np.newaxis])), axis=1)/2**15

        else:  # already normalized
            data = np.concatenate((np.float64(data.real[:, np.newaxis]), np.float64(data.imag[:, np.newaxis])), axis=1)

    elif (abs(data) > 1.0).any():
        data = data / 2**15

    term = 100  # ohms
    vppd = data
    max_adc = np.max(np.abs(vppd), axis=0)
    current = vppd / term  # amps
    p_inst = np.square(current.real)*term # watts
    p_avg_mw = np.sum(p_inst, axis=0)*1000/vppd.size  # mean milliwatts
    dbm_avg = 10*np.log10(p_avg_mw)
    return p_avg_mw, dbm_avg, max_adc


def convert_freq_to_ddc_bins(freqs: Iterable[float | int]) -> np.ndarray:
    """
    Convert frequecnies to bins coming out of the OPFB--useful for programming bin2res
    Args:
        freqs: list or array of frequencies [Hz]

    Returns: np array of bins coming out of the opfb which will contain the frequencies

    """
    bins = np.zeros(2048, dtype=int)
    bins[:freqs.size] = opfb_bin_number(freqs, ssr_raw_order=True)
    return bins


def format_sample_duration(fs: float, n_samp: int) -> str:
    """
    Print the time of some number of samples with a given sample rate in nice units.
    Args:
        fs: sample rate in Hz
        n_samp: number of samples

    Returns:
    For Ex: '0.5 Seconds'

    """
    seconds = n_samp * (1 / fs)

    if seconds > 0.1:
        return f'{seconds:.2f} s'
    elif seconds > 1e-3:
        return f'{seconds / 1e-3:.2f} ms'
    elif seconds > 1e-6:
        return f'{seconds / 1e-6:.2f} us'
    elif seconds > 1e-9:
        return f'{seconds / 1e-9:.2f} ns'


def format_time(t: float) -> str:
    """
    Print the time in a convenient order of magnitude.
    Args:
        t: time in seconds

    Returns:
    For Ex: '0.5 Seconds'

    """

    if t > 0.1:
        return f'{t:.2f} s'
    elif t > 1e-3:
        return f'{t / 1e-3:.2f} ms'
    elif t > 1e-6:
        return f'{t / 1e-6:.2f} us'
    elif t > 1e-9:
        return f'{t / 1e-9:.2f} ns'


def format_bytes(n_bytes: int) -> str:
    """
    Print the number of bytes with a convenient order of magnitude.
    Args:
        n_bytes: number of bytes

    Returns:
    For Ex: '129.2 MiB'
    """
    if n_bytes > 2 ** 30:
        return f'{n_bytes / 2 ** 30:.1f} GiB'
    elif n_bytes > 2 ** 20:
        return f'{n_bytes / 2 ** 20:.1f} MiB'
    elif n_bytes > 2 ** 10:
        return f'{n_bytes / 2 ** 10:.1f} KiB'
    else:
        return f'{n_bytes} bytes'


def delete_pynq_cache_file():
    try:
        os.remove('/usr/local/share/pynq-venv/lib/python3.10/site-packages/pynq/pl_server/_current_metadata.pkl')
    except Exception as e:
        return str(e)


def buf2complex(b, free=True, unsigned=False, floating=True):
    """ Convert a pynq buffer to normal numpy array, copying it out of PL DDR4

    Treat input as uint16 if unsigned is set.

    Divide by 2**15 (or 2**16 if unsigned) if floating is set.

    Frees the pynq buffer if free is set
    """
    x = np.array(b)
    if unsigned:
        x = x.astype(np.uint16, copy=False)
    if floating:
        x /= 2 ** (16 if unsigned else 15)
    x = x[..., 0] + 1j * x[..., 1]
    try:
        if free:
            b.freebuffer()
    except AttributeError:
        pass  # not a pynq buffer
    return x


def setup_logging(name):
    ref = importlib.resources.files('mkidgen3.config').joinpath('logging.yaml')
    with importlib.resources.as_file(ref) as path:
        if os.path.exists(path):
            with open(path, 'rt') as f:
                config = yaml.safe_load(f.read())

    # postprocess loggers dict
    # keys are program names values are either
    #   1) a level for the log of the program
    #   2) a dict of log names and levels
    #   3) a dict of log names and dicts describing how to configure the corresponding Logger instance.
    #  See https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
    # The configuring dict is searched for the following keys:
    #     level (optional). The level of the logger.
    #     propagate (optional). The propagation setting of the logger.
    #     filters (optional). A list of ids of the filters for this logger.
    #     handlers (optional). A list of ids of the handlers for this logger.
    cfg = config['loggers'][name]  # extract one we care about
    if isinstance(cfg, str):
        config['loggers'] = {name: {'level': cfg.upper()}}
    else:
        loggers = {}
        for k, v in cfg.items():
            loggers[k] = {'level': v.upper()} if isinstance(v, str) else v
        config['loggers'] = loggers

    logging.config.dictConfig(config)
    return logging.getLogger(name)


def _which_one_bit_set(x, nbits):
    """
    Given the number x that only has a single bit set return the index of that bit.
    Return None if no bits < nbits bit is set (e.g. nbits=16 will check bits 0-15)
    """
    for i in range(nbits):
        if x & (1 << i):
            return i
    return None


def pack16_to_32(data):
    it = iter(data)
    vals = [x | (y << 16) for x, y in zip(it, it)]
    if data.size % 2:
        vals.append(data[-1])
    return np.array(vals, dtype=np.uint32)


def ensure_array_or_scalar(x):
    if x is None:
        return x
    return x if np.isscalar(x) else np.asarray(x)


def clear_interrupt(core, clear_axi_interrupt_controller=False):
    """
    Attempt to clear an interrupt on both an HLS core and at the Axi interrupt controller. For testing, not production.

    Args:
        core: a pynq.Interrupt or object with a pynq.DefaultIP with both an interrupt and
            register_map.IP_ISR.CHAN0_INT_ST
        clear_axi_interrupt_controller: set to true to reach in and futz with the interrupt controller as well

    Returns: None

    """
    import pynq
    given_int = isinstance(core, pynq.Interrupt)
    interrupt = core if given_int else core.interrupt

    int_number = interrupt.number

    if not given_int and core.register_map.IP_ISR.CHAN0_INT_ST:
        getLogger(__name__).info('Interrupt cleared at controller')

    val = 1 << int_number
    if clear_axi_interrupt_controller:
        parent_mmio = interrupt.parent().mmio
        if not (parent_mmio.read(0x8) & val):
            getLogger(__name__).info('Interrupt is not enabled at controller')
        if parent_mmio.read(0x4) & val:
            parent_mmio.write(0x0C, val)
            getLogger(__name__).info('Interrupt state was cleared at controller')


def ps_ram_sane(nbytes, pynq_region=False):
    """
    Return True if the bytes might be available in PS ram
    Args:
        nbytes: number of bytes
        pynq_region: select if the bytes should be in the region available to pynq.allocate

    Returns: True if the ram might be available

    """
    if pynq_region:
        return nbytes < 2 * 1024 ** 3
    else:
        return nbytes < 2 * 1024 ** 3


def do_asyncio_thing(thing, use_new_thread=False):
    """

    Args:
        thing: an asyncio coroutine. not really clear on the scope that will work safely here yet
           written with do_asyncio_thing(interrupt.wait()) in mind
        use_new_thread: Do the waiting in a new daemon thread

    Returns: None or a future if a new thread is in use

    """
    import asyncio
    from threading import Thread

    def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    newloop = False
    try:
        if use_new_thread:
            raise Exception
        loop = asyncio.get_running_loop()
        getLogger(__name__).debug(f"Using existing loop")
    except:
        loop = asyncio.new_event_loop()
        newloop = True

    if use_new_thread:
        getLogger(__name__).debug(f"Using new loop in new thread")
        thread = Thread(target=start_background_loop, args=(loop,), daemon=True)
        thread.start()
        return asyncio.run_coroutine_threadsafe(thing, loop)

    if newloop:
        getLogger(__name__).debug(f"Using new loop")

    task = loop.create_task(thing)
    getLogger(__name__).debug(f"Will wait in {loop} for {thing} (use_new_thread={use_new_thread})")
    loop.run_until_complete(task)
    if newloop:
        loop.close()


class AbortedException(Exception):
    pass


def check_zmq_abort_pipe(pipe):
    try:
        abort = pipe.recv(zmq.NOBLOCK)
        raise AbortedException(abort)
    except zmq.ZMQError as e:
        if e.errno != zmq.EAGAIN:
            raise


def check_active_jupyter_notebook():
    """Get a list of jupyter notebooks that are running and return true if any have 'http' in the listing """
    x = subprocess.run(['jupyter', 'notebook', 'list'], capture_output=True)
    return False  # 'http' in x.stdout.decode()



import zmq
ctx = zmq.Context.instance()
ctx.linger = 0
import logging
logging.basicConfig(level=logging.DEBUG)
from mkidgen3.server.feedline_config import FeedlineConfig
from mkidgen3.server.captures import CaptureJob, PowerSweepJob, CaptureRequest

# cap command default 8888
# cap data 8889
# cap status 9000
server_ip = 'mkidrfsoc4x2.physics.ucsb.edu'
feedline_server = f'tcp://{server_ip}:8888'
capture_data_server = f'tcp://{server_ip}:8889'
status_server = f'tcp://{server_ip}:8890'

#start a listner for status
# pd = ThreadDevice(zmq.QUEUE, zmq.XSUB, zmq.XPUB)
# pd.bind_in(f'tcp://localhost:9000')
# pd.bind_out('inproc://cap_status')
# # pd.setsockopt_in(zmq.SUBSCRIBE, b"B")
# pd.start()

#start a listner for data

fc = FeedlineConfig()
cr = CaptureRequest(1337, 'iq', fc)
cj = CaptureJob(cr, feedline_server, capture_data_server, status_server, submit=False)
# cj.submit()
#
#
# ps = PowerSweepJob()
# jobs = ps.generate_jobs(submit=True)
# for j in jobs:
#     j.data()   # Will block until data is ready



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import welch
import scipy.optimize as spo
from mkidgen3.server.feedline_config import *
from mkidgen3.server.waveform import WaveformFactory
from mkidgen3.opfb import opfb_bin_number


def fit_psd_floor(freqs, spect, cutoff=30e3):
    def flatoverf(freq, a, b):
        return a + b / freq

    fitcutoff = np.argmin(np.abs(freqs - cutoff))  # remove rolloff
    popt, pcov = spo.curve_fit(flatoverf, freqs[1:fitcutoff], spect[1:fitcutoff], sigma=spect[1:fitcutoff])
    return popt[0]


def get_multi_tone_phase_psd(ro_tone, static_tones, dac_dynamic_range, ol, psd_rolloff_cutoff, program_matched_filt=False, plot=True):
    # Run DAC
    if (np.abs(ro_tone - static_tones) < 1.0e6).any():
        ro_tone = static_tones[np.abs(ro_tone - static_tones) < 1.0e6][0]
        wvfm_tones = static_tones
    else:
        wvfm_tones = np.append(ro_tone, static_tones)
    wvfm_cfg = WaveformConfig(waveform=WaveformFactory(frequencies=wvfm_tones, seed=6, dac_dynamic_range=dac_dynamic_range, compute=True))
    ol.dac_table.configure(**wvfm_cfg.settings_dict())

    # Bin2Res + DDC
    bins = np.zeros(2048, dtype=int)
    bins[0] = opfb_bin_number(ro_tone, ssr_raw_order=True)

    chan = ChannelConfig(bins=bins)
    ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())

    ddc_tones = np.zeros(2048)
    ddc_tones[0] = ro_tone

    ddc = DDCConfig(tones=ddc_tones)
    ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())


    if program_matched_filt:
        # Matched Filter
        filtercfg=FilterConfig(coefficients=f'unity{2048}')
        ol.photon_pipe.phasematch.configure(**filtercfg.settings_dict())

    # Capture Phase
    x = ol.capture.capture_phase(2 ** 19, [0, 1], tap_location='filtphase')
    phase = np.array(x)
    x.freebuffer()
    phase0 = np.pi * phase[:, 0] / (2 ** 15 - 1)

    f, psd = welch(phase0, fs=1e6, nperseg=1e6 / 1e3)

    floor = fit_psd_floor(f, psd, psd_rolloff_cutoff)

    if plot:
        fig, ax = plt.subplots()
        ax.semilogx(f, 10 * np.log10(psd))
        ax.axhline(10 * np.log10(floor),  color='r', linestyle='--')
        ax.set_xlabel(f'Frequency [Hz] ({1e3 * 1e-3:g} kHz resolution)')
        ax.set_ylabel('dBc/Hz')
        ax.grid()
        ax.set_title('Power Spectral Density')

    return f, psd, floor, ro_tone


def get_single_tone_phase_psd(tone, dac_dynamic_range, ol, psd_rolloff_cutoff, program_matched_filt=False, plot=True):
    # Run DAC
    wvfm_cfg = WaveformConfig(waveform=WaveformFactory(frequencies=tone, seed=6, dac_dynamic_range=dac_dynamic_range, compute=True))
    ol.dac_table.configure(**wvfm_cfg.settings_dict())

    # Bin2Res + DDC
    chan = wvfm_cfg.default_channel_config
    ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
    ddc = wvfm_cfg.default_ddc_config
    ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

    if program_matched_filt:
        # Matched Filter
        filtercfg=FilterConfig(coefficients=f'unity{2048}')
        ol.photon_pipe.phasematch.configure(**filtercfg.settings_dict())

    # Capture Phase
    x = ol.capture.capture_phase(2 ** 19, [0, 1], tap_location='filtphase')
    phase = np.array(x)
    x.freebuffer()
    phase0 = np.pi * phase[:, 0] / (2 ** 15 - 1)

    f, psd = welch(phase0, fs=1e6, nperseg=1e6 / 1e3)

    floor = fit_psd_floor(f, psd, psd_rolloff_cutoff)

    if plot:
        fig, ax = plt.subplots()
        ax.semilogx(f, 10 * np.log10(psd))
        ax.axhline(10 * np.log10(floor),  color='r', linestyle='--')
        ax.set_xlabel(f'Frequency [Hz] ({1e3 * 1e-3:g} kHz resolution)')
        ax.set_ylabel('dBc/Hz')
        ax.grid()
        ax.set_title('Power Spectral Density')

    return f, psd, floor


def calculate_psd(data, fs=1e6, fres=1e3, fit_level=True, plot=True):
    plt.figure()
    f, psd = welch(data, fs=fs, nperseg=fs/fres)

    if fit_level:
        pass

    if plot:
        fig, ax = plt.subplots()
        ax.semilogx(f, 10 * np.log10(psd))
        ax.set_xlabel(f'Frequency [Hz] ({fres * 1e-3:g} kHz resolution)')
        ax.set_ylabel('dBc/Hz')
        ax.grid()
        ax.set_title('Power Spectral Density')

    return f, psd





import numpy as np
from mkidgen3.system_parameters import ADC_DAC_INTERFACE_WORD_LENGTH, DAC_RESOLUTION, DAC_LUT_SIZE, N_CHANNELS, \
    SYSTEM_BANDWIDTH, IF_ATTN_STEP, ADC_MAX_V, PHASE_FRACTIONAL_BITS
from typing import Iterable
from mkidgen3.util import ensure_array_or_scalar
import numpy.typing as nt



def convert_adc_raw_to_mv(raw_data: np.ndarray,
                          adc_interface_word_length=ADC_DAC_INTERFACE_WORD_LENGTH,
                          adc_max_v=ADC_MAX_V):
    """
    Args:
        raw_data: integer data
        adc_interface_word_length: max adc int word length
        adc_max_v: max adc v at the SMA input

    Returns:

    """
    return adc_max_v * (raw_data / (2 ** (adc_interface_word_length - 1) - 1)) * 1e6


def raw_iq_to_unit(raw_data: np.ndarray, word_length=ADC_DAC_INTERFACE_WORD_LENGTH):
    """
    Args:
        raw_data: integer data
        word_length: max int word length

    Returns:
        complex iq data on the IQ unit circle
    """
    return raw_data / (2 ** (word_length - 1) - 1)


def raw_phase_to_radian(raw_data: np.ndarray, phase_fractional_bits=PHASE_FRACTIONAL_BITS, scaled=True):
    """
    Args:
        raw_data: integer data fixed point
        phase_fractional_bits: fractional bits in the phase data

    Returns:
        phase data in radians or scaled radians (±pi, not 0 to 2pi)
    """
    x = raw_data.astype(float)
    x[:] /= (2 ** phase_fractional_bits - 1)
    if not scaled:
        x[:] *= np.pi
    return x


def postage_buffer_to_data(buffer: np.ndarray, complex=True, scaled=True):
    """
    convert postage stamp capture data into rids and associated complex values,
    by default omit the first, garbage stamp
    """
    buffer = np.asarray(buffer)  # ensures pulled out of pynqbuffer via copy

    if complex:
        events = buffer[:, 1:, 0] + buffer[:, 1:, 1] * 1j
    else:
        events = buffer[:, 1:]
    if scaled:
        events /= 2 ** 14

    ids = buffer[:, 0, 0].astype(np.uint16)
    return ids, events


def db2lin(values, mode='voltage'):
    """ Convert a value or values in dB to linear units.
    inputs:
    - values: float
        values in dB to be converted to linear units
    -mode: str
        'power': returns dB value in same power units as reference power
        'voltage': returns dB value proportional to RMS voltage
    Example: if 1 dBm is input (that is 1 dB referenced to 1 milliWatt), 'power'
    will return 1.26 mW.
    """
    values = np.asarray(values)
    if mode == 'power':
        return 10 ** (values / 10)
    if mode == 'voltage':
        return 10 ** (values / 20)


def find_relative_amplitudes(if_attenuations):
    """
    Computes the relative amplitudes of the MKID frequencies based on the if_attenuation they looked best at.
    This is used as a helper function to generate the readout waveform with appropriately scaled powers.
    inputs:
    - mkid_frequencies: list of floats
        a list of mkid resonant frequencies
    - if_attenutation: list of floats
        a list of total if_attenuations, one corresponding to each provided mkid frequency in dB
        VALUES ARE ASSUMED TO BE POSITIVE!!
    returns:
    - relative_amplitudes: list of amplitudes between 0 and 1, one for each MKID in linear units.
    """
    return db2lin(-if_attenuations) / db2lin(-if_attenuations).max()


def quantize_frequencies(freqs: Iterable[float | int], rate: float = 4.096e9, n_samples: int = DAC_LUT_SIZE) -> \
        Iterable[float]:
    """
    Quantize frequencies to nearest available value give the sample rate and number of samples.
    Args:
        freqs: frequencies to be quantized
        rate: samples per second [Hz]
        n_samples: number of samples (in the DAC LUT)

    Returns: Quantized frequencies
    """
    freqs = ensure_array_or_scalar(freqs)
    freq_res = rate / n_samples
    return np.asarray(np.round(freqs / freq_res) * freq_res)


def predict_quantization_error(resolution: int = DAC_RESOLUTION, signed: bool = True):
    """
    Predict max quantization error when quantizing to an integer with resolution bits of precision.
    This is effectively the smallest step size allowed by the resolution.
    Args:
        resolution: number of integer bits with which to quantize
        signed: bool

    Returns: Theoretical quantization error assuming values to be quantized have been scaled to maximize dynamic range
             i.e. max value is 2**(resolution-signed) - 1
    Note: It's assumed the quantization step size is small relative to the variation in the signal being quantized.
    """
    max_val = 2 ** (resolution - signed) - 1
    min_val = -2 ** (resolution - signed)
    return (max_val - min_val) / 2 ** resolution


def quantize_to_int(x: Iterable[float | int] | int, resolution: int = DAC_RESOLUTION, signed: bool = True,
                    word_length: bool = ADC_DAC_INTERFACE_WORD_LENGTH,
                    dyn_range: float = 1.0, return_error: bool = True) -> nt.NDArray[int | np.complex64]:
    """
    Scale and quantize values to integers.
    Args:
        x: input value(s)
        resolution: number of bits of precision to quantize with
        signed: whether output values are signed or not
        word_length: n bits in output values, used to compute max allowed value
        dyn_range: what fraction of the DAC dynamic range to use. 1.0 is all
        return_error: return quantization error

    Returns: Quantized value(s)

    """
    assert 0 <= dyn_range <= 1, "dyn_range must be between 0 and 1"
    x = np.asarray(x)
    max_allowed = np.round(dyn_range * (2 ** (resolution - signed) - 1)).astype(int)
    min_allowed = np.round(-dyn_range * (2 ** (resolution - signed))).astype(int)

    if np.iscomplex(x).any():
        max_abs_val = max(x.real.max(), x.imag.max(), np.abs(x.real.min()), np.abs(x.imag.min()))
        y = max_allowed * (x / max_abs_val)  # scale to max allowed int value
        quant_real = np.round(y.real).astype(int)
        quant_imag = np.round(y.imag).astype(int)  # round to int
        quant_real.clip(min_allowed, max_allowed, out=quant_real)
        quant_imag.clip(min_allowed, max_allowed, out=quant_imag)
        error = max((y.real - quant_real).max(), (y.imag - quant_imag).max())
        quant = (quant_real << (word_length - resolution)) + 1j * (quant_imag << (word_length - resolution))

    else:
        max_abs_val = max(x.max(), np.abs(x.min()))
        y = max_allowed * (x / max_abs_val)  # scale to max allowed int value
        quant = np.round(y).astype(int)  # round to int
        quant.clip(min_allowed, max_allowed, out=quant)
        error = (y - quant).max()
        quant <<= (word_length - resolution)
    if return_error:
        return quant, error
    else:
        return quant


def complex_scale(z, max_val):
    """
    Returns complex array rescaled so the maximum real or imaginary value is max_val

    inputs:
    - z: list or array of complex number
        input complex array or list
    - max_val: float
        new maximum real or imaginary value
    """
    ensure_array_or_scalar(z)
    input_max = max(z.real.max(), z.imag.max())
    return max_val * z / input_max


def uniform_freqs(n_channels: int = N_CHANNELS, bandwidth:float | int = SYSTEM_BANDWIDTH, offset: float | int = 0.5e6) -> np.ndarray:
    """
    Returns a list of frequencies corresponding to one per DDC channel, evenly spaced in the bandwidth.
    Args:
        n_channels: Number of channels in the DDC
        bandwidth: Full channelizer bandwidth (ADC Nyquist bandwidth) in Hz
        offset: move tones this far from a bin cetner in Hz

    Returns: See above.
    """
    return np.linspace(-n_channels / 2, n_channels / 2 - 1, n_channels) * (bandwidth / n_channels)+offset


def est_loop_centers(iq):
    """
    Finds the (I,Q) centers of the loops via percentile math
    iq - np.complex array[n_loops, n_samples]
    returns centers[iq.shape[0]]

    see mkidgen2.roach2controls.fitLoopCenters for history
    """
    ictr = (np.percentile(iq.real, 95, axis=1) + np.percentile(iq.real, 5, axis=1)) / 2
    qctr = (np.percentile(iq.imag, 95, axis=1) + np.percentile(iq.imag, 5, axis=1)) / 2

    return ictr + qctr * 1j


"""
def daccomb_old(frequencies, attenuations, phases=None, iq_ratios=None, phase_offsets=None, max_quant_err=.9,
            globalDacAtten=None, lo=None, return_full=True, max_chan=2048, sample_rate=4.096e9, n_iq_bits=32,
            n_samples=2**19):
    
    Creates floating-point DAC frequency comb by adding many complex frequencies together with specified amplitudes and phases.

    The attenuations holds the absolute attenuation for each resonator signal coming out of the DAC.
    Zero attenuation means that the tone amplitude is set to the full dynamic range of the DAC and the
    DAC attenuator(s) are set to 0. Thus, all values in attenuations must be larger than globalDacAtten.
    If you decrease the globalDacAtten, the amplitude in the DAC LUT decreases so that the total
    attenuation of the signal is the same.

    Note: Duplicate frequencies in freqList are ignored when making the comb along with their corresponding attenuation, phase, etc...

    INPUTS:
        frequencies - list of frequencies in the comb.
        attenuations - list of absolute attenuation values (dB) for each frequency.
        phases - list of phases for each complex signal. If None, generates random phases.
        iq_ratios - If None, 50:50 is assumed.
        phase_offsets -
        spike_percentile_limit - loop generateTones() function with random phases to avoid spikes greater than the
        specified percentile in the output comb. Set to >=1 to disable.

    OUTPUTS:
        dictionary with keywords
        I - I(t) values for frequency comb [signed 32-bit integers]
        Q - Q(t)
        quantizedFreqList - list of frequencies after digitial quantiziation
        dacAtten - The global dac hardware attenuation in dB that should be set



    max_quant_err=max(max_quant_err, .01)

    if len(frequencies) != len(attenuations):
        raise ValueError("Need exactly one attenuation value for each resonant frequency!")

    if phases is not None and len(frequencies) != len(phases):
        raise ValueError("Need exactly one phase value for each resonant frequency!")

    if iq_ratios is not None and len(frequencies) != len(iq_ratios):
        raise ValueError("Need exactly one iqRatio value for each resonant frequency!")

    if phase_offsets is not None and len(frequencies) != len(phase_offsets):
        raise ValueError("Need exactly one iqPhaseOffs value for each resonant frequency!")

    if len(frequencies) > max_chan:
        getLogger(__name__).warning(f"Clipping the last {frequencies.size-max_chan}. MAX_CHAN={max_chan}.")
        frequencies = frequencies[:max_chan]
        attenuations = attenuations[:max_chan]
        if phase_offsets is not None:
            phase_offsets=phase_offsets[:max_chan]
        if iq_ratios is not None:
            iq_ratios=iq_ratios[:max_chan]
        if phases is not None:
            phases=phases[:max_chan]

    getLogger(__name__).debug('Generating DAC comb...')

    autoDacAtten = globalDacAtten is None
    if autoDacAtten:
        globalDacAtten = np.amin(attenuations)

    # Calculate relative amplitudes for DAC LUT
    nBitsPerSampleComponent = n_iq_bits / 2
    maxAmp = int(np.round(2 ** (nBitsPerSampleComponent - 1) - 1))  # 1 bit for sign
    amplitudes = maxAmp * 10 ** (-(attenuations - globalDacAtten) / 20)

    # Calculate resonator frequencies for DAC
    LOFreq = parse_lo(lo, frequencies=frequencies, sample_rate=sample_rate)

    dacFreqList = frequencies - LOFreq
    dacFreqList[dacFreqList < 0] += sample_rate  # For +/- freq

    # Make sure dac tones are unique
    dacFreqList, args, args_inv = np.unique(dacFreqList, return_index=True, return_inverse=True)

    rstate = np.random.get_state()
    from numpy.random import MT19937, RandomState, SeedSequence
    np.random.set_state(RandomState(MT19937(SeedSequence(123456789))).get_state())

    # Generate and add up individual tone time series.
    toneDict = generate_dac_comb(dacFreqList, n_samples, sample_rate, return_merged=True,
                                 amplitudes=amplitudes[args], phases=None if phases is None else phases[args],
                                 iq_ratios=None if iq_ratios is None else iq_ratios[args],
                                 phase_offsets=None if phase_offsets is None else phase_offsets[args])

    # This part takes the longest

    iq = toneDict['iq']

    # check that we are utilizing the dynamic range of the DAC correctly
    sig_i = iq.real.std()
    sig_q = iq.imag.std()

    # 10% of the time there should be a point this many sigmas higher than average
    expectedmax_sig = scipy.special.erfinv((iq.size + max_quant_err - 1) / iq.size) * np.sqrt(2)
    if max_quant_err < 1 and sig_i > 0 and sig_q > 0:
        while max(np.abs(iq.real).max() / sig_i, np.abs(iq.imag).max() / sig_q) >= expectedmax_sig:
            getLogger(__name__).warning("The freq comb's relative phases may have added up sub-optimally. "
                                        "Calculating with new random phases")
            toneDict = generate_dac_comb(dacFreqList, n_samples, sample_rate, amplitudes=amplitudes[args], phases=None,
                                         iq_ratios=None if iq_ratios is None else iq_ratios[args],
                                         phase_offsets=None if phase_offsets is None else phase_offsets[args],
                                         return_merged=True)
            iq = toneDict['iq']

    np.random.set_state(rstate)

    dacQuantizedFreqList = (toneDict['frequencies'])[args_inv]
    dacPhaseList = (toneDict['phases'])[args_inv]

    if autoDacAtten:
        highestVal = max(np.abs(iq.real).max(), np.abs(iq.imag).max())
        dBexcess = 20 * np.log10(highestVal / maxAmp)
        dBexcess = np.ceil(4 * dBexcess) / 4  # rounded up to nearest 1/4 dB
        globalDacAtten -= dBexcess
        # reduce to fit into DAC dynamic range and quantize to integer

        if globalDacAtten > 31.75 * 2:
            dB_reduce = globalDacAtten - 31.75 * 2
            getLogger(__name__).warning(f"Unable to fully utilize DAC dynamic range by {dB_reduce} dB")
            globalDacAtten -= dB_reduce
            dBexcess += dB_reduce
        elif np.max((np.abs(iq.real).max(), np.abs(iq.imag).max())) > maxAmp:
            dBexcess += 0.25  # Since there's some rounding there's a small chance we need to decrease by another atten step

        iq /= 10 ** (dBexcess / 20)

    np.round(iq, out=iq)

    highestVal = max(np.abs(iq.real).max(), np.abs(iq.imag).max())

    msg = ('\tGlobal DAC atten: {} dB'.format(globalDacAtten) +
           '\tUsing {} percent of DAC dynamic range\n'.format(highestVal / maxAmp * 100) +
           '\thighest: {} out of {}\n'.format(highestVal, maxAmp) +
           '\tsigma_I: {}  sigma_Q:{}\n'.format(np.std(iq.real), np.std(iq.imag)) +
           '\tLargest val_I: {} sigma. '.format(np.abs(iq.real).max() / np.std(iq.real)) +
           'val_Q: {} sigma.\n'.format(np.abs(iq.imag).max() / np.std(iq.imag)) +
           '\tExpected val: {} sigmas\n'.format(expectedmax_sig))
    getLogger(__name__).debug(msg)

    if globalDacAtten < 0:
        raise ValueError("Desired resonator powers are unacheivable. "
                         f"Increase resonator attens by {-1 * globalDacAtten} dB")

    if return_full:
        return {'frequencies': dacQuantizedFreqList, 'attenuation': globalDacAtten,
                'comb': iq, 'phases': dacPhaseList}
    else:
        return iq
"""


from pynq import DefaultIP


class AxisSwitch(DefaultIP):
    bindto = ['xilinx.com:ip:axis_switch:1.1']

    def __init__(self, description):
        super().__init__(description=description)

    def set_driver(self, slave=0, master=0, disable=False, commit=True):
        """Set the slave for the master, committing the change triggers a 16 cycle soft reset of the switch"""
        cfg = slave & 0b111
        master = min(max(int(master), 0), 15)
        if disable:
            cfg |= 0x80000000
        self.write(0x0040 + master * 4, cfg)
        if commit:
            self.commit()

    def disable(self, master=0):
        """Disable the master, do not change the driver, will commit and soft-reset the core"""
        master = min(max(int(master), 0), 15)
        addr = 0x0040 + master * 4
        self.write(addr, self.read(addr) | 0x80000000)
        self.commit()

    def enable(self, master=0):
        """Enable the master, do not change the driver, will commit and soft-reset the core"""
        master = min(max(int(master), 0), 15)
        addr = 0x0040 + master * 4
        self.write(addr, self.read(addr) ^ 0x80000000)
        self.commit()

    def is_disabled(self, master=0):
        """True iff the master is disabled"""
        return (self.read(0x0040 + min(max(int(master), 0), 15) * 4) & 0xf0000000) != 0

    def driver_of(self, master=0):
        """Return the driver of the specified master, master may be disabled"""
        return self.read(0x0040 + min(max(int(master), 0), 15) * 4)

    def commit(self):
        """Commit config, triggers a soft 16 cycle reset"""
        self.write(0x0000, 0x2)


class SwitchOnLast(DefaultIP):
    bindto = ['mazinlab:mkidgen3:switch_on_last:0.1']

    def __init__(self, description):
        super().__init__(description=description)

    def set_driver(self, slave=0, disable=False):
        """Set the slave for the output"""
        slave = max(min(slave, 5), 0)
        if disable:
            self.register_map.enable = False
        self.register_map.stream = slave

    def disable(self):
        """Disable the output"""
        self.register_map.enable = False

    def enable(self):
        """Enable the output"""
        self.register_map.enable = True

    def is_disabled(self):
        """True iff the output is disabled"""
        return self.register_map.enable.enable

    def driver_of(self):
        """Return the driver for the output, output may be disabled"""
        return self.register_map.stream.stream



import psutil
import blosc2
from mkidgen3.system_parameters import PL_TOTAL_BYTES, SYSTEM_OVERHEAD_BYTES, COMPRESSION_OVERHEAD_BYTES


def memfree_mib():
    return psutil.virtual_memory().available//1024**2


def determine_max_chunk(domain: str, assume_compression=False, demands: list | tuple | None = None) -> int:
    """
    Attempt to estimate how big of a chunk of memory can be allocated w/o swapping.

    Args:
        domain: Are we working in the PS or the PL?
        demands: An iterable of numbers containing existing pressures on the memory
        assume_compression (False) include enough memory for compression of the data in the ps side

    Returns: a number of

    """
    domain = domain.lower()
    if domain not in ('ps', 'pl'):
        raise ValueError('Valid domains are ps and pl')

    total = psutil.virtual_memory().total if domain == 'ps' else PL_TOTAL_BYTES

    if domain == 'ps':
        total -= SYSTEM_OVERHEAD_BYTES

    demands = list(demands) if demands else []

    if assume_compression:
        demands.append(blosc2.MAX_OVERHEAD)

    left = int(total - sum(demands))

    if assume_compression:
        left = min(blosc2.MAX_BUFFERSIZE, left)

    return left



try:
    from . import axiswitch, bintores, capture, ddc, dactable, axififo, phasematch, iqgen, rfdc, trigger, ppssync
except OSError:
    pass
__all__ = ['axiswitch', 'bintores', 'capture', 'ddc', 'dactable', 'axififo', 'rfdc', 'iqgen',
           'phasematch', 'trigger', 'ppssync']



import asyncio
import copy
from mkidgen3.rfsocmemory import determine_max_chunk, memfree_mib
import mkidgen3.drivers.rfdcclock
import mkidgen3.drivers.rfdc
from pynq import Overlay
from logging import getLogger
from mkidgen3.mkidpynq import DummyOverlay
from mkidgen3.system_parameters import ADC_MAX_INT
from mkidgen3.drivers.ifboard import IFBoard
from mkidgen3.server.feedline_config import (FeedlineConfig, FeedlineConfigManager,
                                             BitstreamConfig, RFDCClockingConfig, RFDCConfig)
from mkidgen3.server.captures import CaptureRequest
from mkidgen3.util import check_zmq_abort_pipe, AbortedException, format_bytes, compute_max_val, format_time
from ..interrupts import ThreadedPLInterruptManager
import zmq
from mkidgen3.server.misc import zpipe
import threading
import pynq
import time
import numpy as np
from mkidgen3.mkidpynq import unpack_photons
DEFAULT_BIT_FILE = '/home/xilinx/gen3_top_final.bit'


class FeedlineHardware:
    def __init__(self, bitstream=DEFAULT_BIT_FILE, rfdcclock='4.096GSPS_MTS_dualloop', if_port='dev/ifboard',
                 ignore_version=False,
                 program_clock=True, rfdc: RFDCConfig | None = None, download=False, clock_source='external'):
        """

        Args:
            bitstream:
            rfdcclock:
            if_port:
            ignore_version:
            program_clock:
            rfdc: Enable MTS. Cannot be enabled unless until overlay is downloaded.
            download:
            clock_source:
        """

        self._default_bitstream = BitstreamConfig(bitstream=bitstream, ignore_version=ignore_version)
        self._default_rfdc_clocking = RFDCClockingConfig(programming_key=rfdcclock, clock_source=clock_source)
        self._default_rfdc = RFDCConfig(dac_mts=True, adc_mts=True) if rfdc is None else rfdc

        self.config_manager = FeedlineConfigManager()

        if program_clock:
            mkidgen3.drivers.rfdcclock.configure(**self._default_rfdc_clocking.settings_dict())
            time.sleep(0.5)  # allow clocks to stabilize before loading overlay  TODO: is this necessary

        try:
            self._ol = pynq.Overlay(self._default_bitstream.bitstream,
                                    ignore_version=self._default_bitstream.ignore_version, download=download)
            mkidgen3.quirks.Overlay(self._ol).post_configure()

            if download:
                ThreadedPLInterruptManager.get_manager()
                self._ol.rfdc.enable_mts(dac=self._default_rfdc.dac_mts, adc=self._default_rfdc.adc_mts)

        except RuntimeError as e:
            if 'No Devices Found' in str(e):
                getLogger(__name__).warning('No PL device found, is BOARD set? This is expected on a laptop')
                self._ol = DummyOverlay(bitstream)
            else:
                raise

        self._if_board = IFBoard(if_port, connect=False)

    def reset(self):
        raise NotImplementedError('Reset settings not implemented')
        self._if_board.power_off(save_settings=False)
        mkidgen3.drivers.rfdcclock.configure(self._clock_source)
        self._ol = Overlay(self._ol.bitfile_name, ignore_version=self._ignore_version, download=True)
        self._if_board.power_on()

    def status(self):
        """

        Returns: a JSON Serializable status object per the schema fully describing the state of the feedline hardware

        """
        return 'hardware status'

    def bequiet(self, stop_dacs=True, poweroff_if=False):
        """

        Args:
            stop_dacs: Stop the DACs from replaying any values
            poweroff_if: Power down the IF board (implies `stop_if`)

        Returns: None
        """
        if stop_dacs:
            self._ol.dac_table.quiet()
        if poweroff_if:
            self._if_board.power_off(save_settings=False)

    def config_compatible_with(self, config: FeedlineConfig):
        return self.config_manager.required().compatible_with(config)

    def derequire_config(self, id):
        """True iff the required settings changed as a result"""
        try:
            return self.config_manager.pop(id)
        except KeyError:
            return False

    def apply_config(self, id, config: FeedlineConfig):
        """

        Args:
            id: an identifier to associate with the config, identifiers should be unique to each config.
            config: a feedline config to apply, only updates will actually be sent to hardware

        Returns: None

        Raises: ValueError may be raised if a non-unique ID is used. ValueError will be raised if the config is not
        compatible with currently required configs or it contains settings that are hashed and
        have not been seen before.

        """

        # Add the config to the pot and get the effective config
        fl_setup = self.config_manager.add(id, config)

        clock_set = False
        if fl_setup.rfdc_clk:
            getLogger(__name__).debug(f'Requesting update to RFDC clocking configuration.')
            mkidgen3.drivers.rfdcclock.configure(**fl_setup.rfdc_clk.settings_dict())
            #time.sleep(0.5)
            clock_set = True

        if clock_set or fl_setup.bitstream:
            getLogger(__name__).debug(f'Requesting update to bitstream.')
            if fl_setup.bitstream:
                x = copy.deepcopy(self._default_bitstream)
                x.merge_with(fl_setup.bitstream)
            else:
                x = self._default_bitstream

            self._ol = Overlay(x.bitstream, ignore_version=x.ignore_version, download=True)
            mkidgen3.quirks.Overlay(self._ol).post_configure()
            ThreadedPLInterruptManager.get_manager()

        if fl_setup.rfdc:
            getLogger(__name__).debug(f'Requesting update to RFDC configuration.')
            self._ol.rfdc.enable_mts(dac=fl_setup.rfdc.dac_mts, adc=fl_setup.rfdc.adc_mts)
            self._ol.rfdc.set_gain(adc_gains=fl_setup.rfdc.adc_gains, dac_gains=fl_setup.rfdc.dac_gains)

        from mkidgen3.drivers.ppssync import PPSMode, PPSSource
        self._ol.pps_synchronization.pps_synchronizer_con_0.start_engine(
                mode=PPSMode.FORCE_START,
                start_second=None,
                skew=None,
                lockout=None,
                rollover_thresh=None,
                resync=False,
                pps_source=PPSSource.PPS0,
                load_time=None,
                clk_period_ns=1.953125,
                timeout=5,
                poll=1000*1000*1000)

        #NB these if comparisons must not be "is None" as an empty config can't be used to trigger reconfiguration

        if fl_setup.if_board:
            getLogger(__name__).debug(f'Requesting update to IF Board configuration.')
            try:
                self._if_board.configure(**fl_setup.if_board.settings_dict())
            except Exception as e:
                getLogger(__name__).warning(f"Unable to connect to IF board: {e}, proceeding without IF board.")

        if fl_setup.waveform:
            getLogger(__name__).debug(f'Configure DAC with {fl_setup.waveform.settings_dict()}')
            self._ol.dac_table.configure(**fl_setup.waveform.settings_dict())

        if fl_setup.chan:
            self._ol.photon_pipe.reschan.bin_to_res.configure(**fl_setup.chan.settings_dict())

        if fl_setup.ddc:
            self._ol.photon_pipe.reschan.ddccontrol_0.configure(**fl_setup.ddc.settings_dict())

        if fl_setup.filter:
            self._ol.photon_pipe.phasematch.configure(**fl_setup.filter.settings_dict())

        if fl_setup.trig:
            self._ol.trigger_system.trigger_1.configure(**fl_setup.trig.settings_dict())

        if fl_setup.if_board:
            self._if_board.settle()

    def plram_cap(self, pipe, cr: CaptureRequest, context=None):
        """

        Args:
            pipe: An inproc
            context:
            cr: A CaptureRequest object

        Returns: None

        """

        failmsg = ''
        try:
            assert cr.type == 'engineering', 'Incorrect capture request type'
            assert self._ol.capture.is_ready(), 'Capture Subsystem is busy'
        except AssertionError as e:
            failmsg = str(e)
        except AttributeError as e:
            failmsg = f'Something is fundamentally wrong. Check your bitstream. ({str(e)})'

        try:
            cr.establish(context=context)
        except zmq.ZMQError as e:
            failmsg += f"Unable to establish capture {cr.id} due to {e}, dropping request."

        if not failmsg and cr.fail_saturation_frac:  # check for ADC saturation
            buf = self._ol.capture.capture_adc(2**19, complex=False)
            max_val = np.abs(buf).max()
            del buf
            if max_val >= ADC_MAX_INT*cr.fail_saturation_frac:
                failmsg = (f"ADC levels ({max_val}) above saturation failure "
                           f"level ({ADC_MAX_INT*cr.fail_saturation_frac:.0f})")

        if failmsg:
            getLogger(__name__).error(failmsg)
            cr.fail(failmsg, raise_exception=False)
            try:
                pipe.close()
            except zmq.ZMQError:
                pass
            return

        self._ol.capture.keep_channels(cr.tap, cr.channels if cr.channels else 'all')
        hw_channels = tuple(sorted(self._ol.capture.kept_channels(cr.tap)))

        capture_atom_bytes = len(hw_channels)*cr.dwid
        hw_size_bytes = capture_atom_bytes*cr.nsamp
        demands = None
        chunking_thresh = determine_max_chunk('ps', demands=demands,
                                              assume_compression=not cr.data_endpoint.startswith('file://'))
        nchunks = hw_size_bytes // chunking_thresh
        partial = hw_size_bytes - chunking_thresh * nchunks
        chunks = [chunking_thresh // capture_atom_bytes] * nchunks
        if partial:
            chunks.append(partial // capture_atom_bytes)

        if len(chunks) > 1 and cr.numpy_metric:
            failmsg = ('Reducing captures via numpy metrics is not supported chunked captures. '
                       'Decrease the capture size or compute the metric client-side.')
            getLogger(__name__).error(failmsg)
            cr.fail(failmsg, raise_exception=False)
            try:
                pipe.close()
            except zmq.ZMQError:
                pass
            return

        channel_sel = None
        ps_buf = None
        if cr.channels is not None:  # channel-specific capture
            usr_channels = cr.channels
            if hw_channels != usr_channels:  # strip off extra channels required by hardware capture
                getLogger(__name__).debug(f'Requested channel capture of {format_bytes(cr.size_bytes)} '
                                          f'requires {format_bytes(hw_size_bytes)}')
                getLogger(__name__).debug(f'Requested channel subset of hardware capture '
                                              f'will copy {format_bytes(cr.size_bytes)} '
                                              f'PL buffer to PS RAM.')
                channel_sel = np.where(np.in1d(hw_channels, usr_channels))[0]
                buf_shape = (chunks[0], len(usr_channels))
                if 'iq' in cr.tap:
                    buf_shape += (2,)
                ps_buf = np.empty(buf_shape,  dtype=np.int16)

        getLogger(__name__).debug(f'Beginning plram capture loop of {len(chunks)} chunk(s) at {cr.tap}')

        try:
            for i, csize in enumerate(chunks):
                times = [time.perf_counter()]

                check_zmq_abort_pipe(pipe)

                times.append(time.perf_counter())
                data = self._ol.capture.capture(csize, cr.tap, groups=None, wait=True)
                times.append(time.perf_counter())
                if isinstance(data, dict):
                    raise RuntimeError(f'PL axis2mm capture failed: {data}. '
                                       f'Note decode errors are indicative of a memory address and '
                                       f'should never happen issue.')
                if channel_sel is None:
                    data_to_send = data
                else:
                    data_to_send = np.take(data, channel_sel, axis=1, out=ps_buf[:csize])
                times.append(time.perf_counter())

                zmqtmp = zmq.COPY_THRESHOLD
                zmq.COPY_THRESHOLD = 0
                tracker = cr.send_data(data_to_send, status=f'{i + 1}/{len(chunks)}', copy=False, compress=True)
                times.append(time.perf_counter())

                if tracker is not None:
                    tracker.wait()
                zmq.COPY_THRESHOLD = zmqtmp
                times.append(time.perf_counter())

                del data
                times.append(time.perf_counter())
                times_str = [format_time(x) for x in np.diff((times))]
                labels = ['check abort','execute capture','optionally slice','send data','wait for tracker','free buffer']
                getLogger(__name__).debug(list(zip(labels, times_str)))
            cr.finish()
        except AbortedException as e:
            cr.abort(e)
        except Exception as e:
            getLogger(__name__).error(f'Terminating {cr} due to {e}')
            cr.fail(f'Aborted due to {e}', raise_exception=False)
        finally:
            cr.destablish()
            del ps_buf
            getLogger(__name__).debug(f'plram cap complete')

    def photon_cap(self, pipe: zmq.Socket, cr: CaptureRequest, context=None):
        """
        pipe: a zmq pair pipe to detect abort
        cr: the capture request
        """
        failmsg = ''
        photon_maxi = self._ol.trigger_system.photon_maxi_0
        try:
            assert cr.type == 'photon', 'Incorrect capture request type'
        except AssertionError as e:
            failmsg = str(e)

        try:
            cr.establish(context=context)
        except zmq.ZMQError as e:
            failmsg = f"Unable to establish capture {cr.id} due to {e}, dropping request."

        if failmsg:
            getLogger(__name__).error(failmsg)
            try:
                cr.fail(failmsg)
                cr.destablish()
            except zmq.ZMQError as ez:
                getLogger(__name__).warning(f'Failed to send abort/destablish for {cr} due to {ez}')
            return

        q, q_other = zpipe(zmq.Context.instance())
        # q = q_other = queue.SimpleQueue()  #an alternative

        fountain, stop = photon_maxi.photon_fountain(q_other, spawn=True, copy_buffer=True)

        def photon_sender(q: zmq.Socket, cr, unpack=False):
            log = getLogger(__name__)
            from datetime import datetime
            toc = 0
            delts = []
            try:
                cr.establish(context=context)
                while True:
                    tic = datetime.utcnow()
                    if toc:
                        delt = tic - toc
                        delts.append(delt.seconds*1e6+delt.microseconds)
                        x = tic.strftime('%M:%S.%f')
                        log.debug(f'Prep send @ {x}, since last wait ended {delt.seconds}.{delt.microseconds:06}s')

                    x = q.recv()
                    if x == b'':
                        toc = datetime.utcnow()
                        cr.finish()
                        break
                    photons = np.frombuffer(x, dtype=photon_maxi.PHOTON_PACKED_DTYPE)
                    toc = datetime.utcnow()

                    cr.send_data(unpack_photons(photons) if unpack else photons, copy=False, compress=True)
            except Exception as e:
                cr.abort(f'Uncaught exception in photon sender: {e}')
                raise e
            finally:
                cr.destablish()
                q.close()
            log.info(f'Sending done. Time between send waits: {delts} us')

        cr.destablish()
        sender = threading.Thread(target=photon_sender, args=(q, cr), name='Photon Sender')

        try:
            sender.start()
            fountain.start()
            time.sleep(.2)
            photon_maxi.capture(buffer_time_ms=cr.nsamp)
            while not cr.completed:
                check_zmq_abort_pipe(pipe)
        except AbortedException as e:
            getLogger(__name__).error(f'Aborting photon capture {cr} due user request.')
            stop.set()
        except Exception as e:
            getLogger(__name__).error(f'Terminating {cr} due to {e}')
            stop.set()
        finally:
            self._ol.trigger_system.photon_maxi_0.stop_capture()
            fountain.join()
            sender.join()
            getLogger(__name__).debug(f'Deleting {cr} as all work is complete')
            del cr
            if isinstance(q, zmq.Socket):
                q.close()
                q_other.close()

    def postage_cap(self, pipe: zmq.Socket, cr: CaptureRequest, context: zmq.Context = None):
        failmsg = ''
        postage_filt = self._ol.trigger_system.postage_filter_0
        postage_maxi = self._ol.trigger_system.postage_maxi_0
        try:
            assert cr.type == 'postage', 'Incorrect capture request type'
            assert postage_maxi.register_map.CTRL.AP_IDLE == 1, 'Postage MAXI is busy'
        except AssertionError as e:
            failmsg = str(e)

        try:
            cr.establish(context=context)
        except zmq.ZMQError as e:
            failmsg = f"Unable to establish capture {cr.id} due to {e}, dropping request."

        if failmsg:
            getLogger(__name__).error(failmsg)
            cr.fail(failmsg, raise_exception=False)
            return

        try:
            postage_filt.configure(monitor_channels=cr.channels)
            _, postdone = ThreadedPLInterruptManager.get_monitor(postage_maxi, id='frs_postage_cap')
            postdone.clear()
            postage_maxi.capture(max_events=cr.nsamp, wait=False)

            while not postdone.is_set():
                check_zmq_abort_pipe(pipe)
                time.sleep(min(postage_maxi.MAX_CAPTURE_TIME_S / 10, .1))
            cr.send_data(postage_maxi.get_postage(rawbuffer=True), copy=False, compress=True)
            cr.finish()
        except AbortedException as e:
            cr.abort(e, raise_exception=False)
        except Exception as e:
            getLogger(__name__).error(f'Terminating {cr} due to {e}')
            cr.fail(f'Failed due to {e}', raise_exception=False)
        finally:
            getLogger(__name__).debug(f'Deleting {cr} as all work is complete')
            del cr



import numpy as np
import scipy.special
from logging import getLogger
import logging
from mkidgen3.gen2 import SweepFile, parse_lo
from mkidgen3.opfb import quantize_frequencies

DAC_REPLAY_SAMPLES=262144


def generate_dac_comb(frequencies, n_samples, sample_rate, amplitudes=None, phases=None, iq_ratios=None,
                      phase_offsets=None, return_merged=True):
    """
    Generate a list of complex signals with amplitudes and phases specified and frequencies quantized

    All specified inputs must be of the same shape

    INPUTS:
        frequencies - resonator frequencies
        n_samples - Number of time samples
        sample_rate - Used to quantize the frequencies
        amplitudes - Tone amplitudes (0 -1). If None, use 1.
        phases - Phases to use. If None, use random phase
        return_merged - set to false to return timeseries for each frequency independently. Uses frequencies.size times
        more memory

    OUTPUTS (in dict):
        iq - complex64 array of IQ(t) values. If unmerged will have an additional axis corresponding to frequency
        frequencies - quantized frequencies
        phases - phases for each frequency
    """
    if amplitudes is None:
        amplitudes = np.ones_like(frequencies)
    if phases is None:
        phases = np.random.uniform(0., 2. * np.pi, len(frequencies))
    if iq_ratios is None:
        iq_ratios = np.ones_like(frequencies)
    if phase_offsets is None:
        phase_offsets = np.zeros_like(frequencies)

    # Quantize the frequencies to their closest digital value
    quantized_freqs = quantize_frequencies(frequencies, rate=sample_rate, n_samples=n_samples)
    phase_offsets_radians = np.deg2rad(phase_offsets)

    if return_merged:
        iq = np.zeros(n_samples, dtype=np.complex64)
    else:
        iq = np.zeros((frequencies.size, n_samples), dtype=np.complex64)

    # generate each signal
    t = 2 * np.pi * np.arange(n_samples)/sample_rate
    for i in range(frequencies.size):
        phi = t * quantized_freqs[i]
        exp = amplitudes[i] * np.exp(1j * (phi + phases[i]))
        iScale = np.sqrt(2) * iq_ratios[i] / np.sqrt(1. + iq_ratios[i] ** 2)
        qScale = np.sqrt(2) / np.sqrt(1 + iq_ratios[i] ** 2)
        if return_merged:
            iq.real += iScale * (np.cos(phase_offsets_radians[i]) * np.real(exp) +
                                 np.sin(phase_offsets_radians[i]) * np.imag(exp))
            iq.imag += qScale * np.imag(exp)
        else:
            iq[i].real = iScale * (np.cos(phase_offsets_radians[i]) * np.real(exp) +
                                   np.sin(phase_offsets_radians[i]) * np.imag(exp))
            iq[i].imag = qScale * np.imag(exp)

    return {'iq': iq, 'frequencies': quantized_freqs, 'phases': phases}


def daccomb(frequencies, attenuations, phases=None, iq_ratios=None, phase_offsets=None, spike_percentile_limit=.9,
             globalDacAtten=None, lo=None, return_full=True, max_chan=2048, sample_rate=4.096e9, n_iq_bits=32,
             n_samples=2**19):
    """
    Creates DAC frequency comb by adding many complex frequencies together with specified amplitudes and phases.

    The attenuations holds the absolute attenuation for each resonator signal coming out of the DAC.
    Zero attenuation means that the tone amplitude is set to the full dynamic range of the DAC and the
    DAC attenuator(s) are set to 0. Thus, all values in attenuations must be larger than globalDacAtten.
    If you decrease the globalDacAtten, the amplitude in the DAC LUT decreases so that the total
    attenuation of the signal is the same.

    Note: The freqList need not be unique. If there are repeated values in the freqList then
    they are completely ignored when making the comb along with their corresponding attenuation, phase, etc...

    INPUTS:
        frequencies - list of all resonator frequencies.
        attenuations - list of absolute attenuation values (dB) for each resonator.
        phases - list of phases for each complex signal. If None, generates random phases.
        iq_ratios -
        phase_offsets -
        spike_percentile_limit - loop generateTones() function with random phases to avoid spikes greater than the
        specified percentile in the output comb. Set to >=1 to disable.

    OUTPUTS:
        dictionary with keywords
        I - I(t) values for frequency comb [signed 32-bit integers]
        Q - Q(t)
        quantizedFreqList - list of frequencies after digitial quantiziation
        dacAtten - The global dac hardware attenuation in dB that should be set

    """

    spike_percentile_limit=max(spike_percentile_limit, .01)

    if len(frequencies) != len(attenuations):
        raise ValueError("Need exactly one attenuation value for each resonant frequency!")

    if phases is not None and len(frequencies) != len(phases):
        raise ValueError("Need exactly one phase value for each resonant frequency!")

    if iq_ratios is not None and len(frequencies) != len(iq_ratios):
        raise ValueError("Need exactly one iqRatio value for each resonant frequency!")

    if phase_offsets is not None and len(frequencies) != len(phase_offsets):
        raise ValueError("Need exactly one iqPhaseOffs value for each resonant frequency!")

    if len(frequencies) > max_chan:
        getLogger(__name__).warning(f"Clipping the last {frequencies.size-max_chan}. MAX_CHAN={max_chan}.")
        frequencies = frequencies[:max_chan]
        attenuations = attenuations[:max_chan]
        if phase_offsets is not None:
            phase_offsets=phase_offsets[:max_chan]
        if iq_ratios is not None:
            iq_ratios=iq_ratios[:max_chan]
        if phases is not None:
            phases=phases[:max_chan]

    getLogger(__name__).debug('Generating DAC comb...')

    autoDacAtten = globalDacAtten is None
    if autoDacAtten:
        globalDacAtten = np.amin(attenuations)

    # Calculate relative amplitudes for DAC LUT
    nBitsPerSampleComponent = n_iq_bits / 2
    maxAmp = int(np.round(2 ** (nBitsPerSampleComponent - 1) - 1))  # 1 bit for sign
    amplitudes = maxAmp * 10 ** (-(attenuations - globalDacAtten) / 20)

    # Calculate resonator frequencies for DAC
    LOFreq = parse_lo(lo, frequencies=frequencies, sample_rate=sample_rate)

    dacFreqList = frequencies - LOFreq
    dacFreqList[dacFreqList < 0] += sample_rate  # For +/- freq

    # Make sure dac tones are unique
    dacFreqList, args, args_inv = np.unique(dacFreqList, return_index=True, return_inverse=True)

    rstate = np.random.get_state()
    from numpy.random import MT19937, RandomState, SeedSequence
    np.random.set_state(RandomState(MT19937(SeedSequence(123456789))).get_state())

    # Generate and add up individual tone time series.
    toneDict = generate_dac_comb(dacFreqList, n_samples, sample_rate, return_merged=True,
                                 amplitudes=amplitudes[args], phases=None if phases is None else phases[args],
                                 iq_ratios=None if iq_ratios is None else iq_ratios[args],
                                 phase_offsets=None if phase_offsets is None else phase_offsets[args])

    # This part takes the longest

    iq = toneDict['iq']

    # check that we are utilizing the dynamic range of the DAC correctly
    sig_i = iq.real.std()
    sig_q = iq.imag.std()

    # 10% of the time there should be a point this many sigmas higher than average
    expectedmax_sig = scipy.special.erfinv((iq.size + spike_percentile_limit - 1)/ iq.size) * np.sqrt(2)
    if spike_percentile_limit < 1 and sig_i > 0 and sig_q > 0:
        while max(np.abs(iq.real).max() / sig_i, np.abs(iq.imag).max() / sig_q) >= expectedmax_sig:
            getLogger(__name__).warning("The freq comb's relative phases may have added up sub-optimally. "
                                        "Calculating with new random phases")
            toneDict = generate_dac_comb(dacFreqList, n_samples, sample_rate, amplitudes=amplitudes[args], phases=None,
                                         iq_ratios=None if iq_ratios is None else iq_ratios[args],
                                         phase_offsets=None if phase_offsets is None else phase_offsets[args],
                                         return_merged=True)
            iq = toneDict['iq']

    np.random.set_state(rstate)

    dacQuantizedFreqList = (toneDict['frequencies'])[args_inv]
    dacPhaseList = (toneDict['phases'])[args_inv]

    if autoDacAtten:
        highestVal = max(np.abs(iq.real).max(), np.abs(iq.imag).max())
        dBexcess = 20 * np.log10(highestVal / maxAmp)
        dBexcess = np.ceil(4 * dBexcess) / 4  # rounded up to nearest 1/4 dB
        globalDacAtten -= dBexcess
        # reduce to fit into DAC dynamic range and quantize to integer

        if globalDacAtten > 31.75 * 2:
            dB_reduce = globalDacAtten - 31.75 * 2
            getLogger(__name__).warning(f"Unable to fully utilize DAC dynamic range by {dB_reduce} dB")
            globalDacAtten -= dB_reduce
            dBexcess += dB_reduce
        elif np.max((np.abs(iq.real).max(), np.abs(iq.imag).max())) > maxAmp:
            dBexcess += 0.25  # Since there's some rounding there's a small chance we need to decrease by another atten step

        iq /= 10 ** (dBexcess / 20)

    np.round(iq, out=iq)

    highestVal = max(np.abs(iq.real).max(), np.abs(iq.imag).max())

    msg = ('\tGlobal DAC atten: {} dB'.format(globalDacAtten) +
           '\tUsing {} percent of DAC dynamic range\n'.format(highestVal / maxAmp * 100) +
           '\thighest: {} out of {}\n'.format(highestVal, maxAmp) +
           '\tsigma_I: {}  sigma_Q:{}\n'.format(np.std(iq.real), np.std(iq.imag)) +
           '\tLargest val_I: {} sigma. '.format(np.abs(iq.real).max() / np.std(iq.real)) +
           'val_Q: {} sigma.\n'.format(np.abs(iq.imag).max() / np.std(iq.imag)) +
           '\tExpected val: {} sigmas\n'.format(expectedmax_sig))
    getLogger(__name__).debug(msg)

    if globalDacAtten < 0:
        raise ValueError("Desired resonator powers are unacheivable. "
                         f"Increase resonator attens by {-1 * globalDacAtten} dB")

    if return_full:
        return {'frequencies': dacQuantizedFreqList, 'attenuation': globalDacAtten,
                'comb': iq, 'phases': dacPhaseList}
    else:
        return iq


def meccomb(mec_freqfile, lo):
    freqfile = SweepFile(mec_freqfile)
    combdata = daccomb(frequencies=freqfile.freq, attenuations=freqfile.atten, phases=freqfile.phases,
                       iq_ratios=freqfile.iqRatios, globalDacAtten=None, lo=lo)
    return combdata
